@@ -114,7 +114,7 @@ oom:
     goto cleanup;
 }
 
-int capabilities_print_json(bool cusetpm)
+int capabilities_print_json(bool cusetpm, TPMLIB_TPMVersion tpmversion)
 {
     char *string = NULL;
     int ret = -1;
@@ -130,6 +130,9 @@ int capabilities_print_json(bool cusetpm)
     const char *nvram_backend_dir = "\"nvram-backend-dir\", ";
     const char *nvram_backend_file = "\"nvram-backend-file\"";
 
+    /* ignore errors */
+    TPMLIB_ChooseTPMVersion(tpmversion);
+
     ret = get_rsa_keysize_caps(&keysizecaps);
     if (ret < 0)
         goto cleanup;
@@ -143,18 +146,22 @@ int capabilities_print_json(bool cusetpm)
          "{ "
          "\"type\": \"swtpm\", "
          "\"features\": [ "
-             "%s%s%s%s%s%s%s%s%s%s%s"
+             "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s"
           " ], "
          "\"version\": \"" VERSION "\" "
          "}",
          with_tpm1,
          with_tpm2,
          !cusetpm     ? "\"tpm-send-command-header\", ": "",
-         !cusetpm     ? "\"flags-opt-startup\", "      : "",
+         true         ? "\"flags-opt-startup\", "      : "",
+         true         ? "\"flags-opt-disable-auto-shutdown\", ": "",
+         true         ? "\"ctrl-opt-terminate\", "     : "",
          cmdarg_seccomp,
          true         ? "\"cmdarg-key-fd\", "          : "",
          true         ? "\"cmdarg-pwd-fd\", "          : "",
          true         ? "\"cmdarg-print-states\", "    : "",
+         true         ? "\"cmdarg-chroot\", "          : "",
+         true         ? "\"cmdarg-migration\", "       : "",
          nvram_backend_dir,
          nvram_backend_file,
          keysizecaps  ? keysizecaps                    : ""
