@@ -70,6 +70,7 @@
 #include "server.h"
 #include "seccomp_profile.h"
 #include "tpmlib.h"
+#include "mainloop.h"
 
 /* --log %s */
 static const OptionDesc logging_opt_desc[] = {
@@ -127,8 +128,7 @@ static const OptionDesc pid_opt_desc[] = {
     {
         .name = "file",
         .type = OPT_TYPE_STRING,
-    },
-    {
+    }, {
         .name = "fd",
         .type = OPT_TYPE_INT,
     },
@@ -140,12 +140,10 @@ static const OptionDesc tpmstate_opt_desc[] = {
     {
         .name = "dir",
         .type = OPT_TYPE_STRING,
-    },
-    {
+    }, {
         .name = "mode",
         .type = OPT_TYPE_MODE_T,
-    },
-    {
+    }, {
         .name = "backend-uri",
         .type = OPT_TYPE_STRING,
     },
@@ -183,6 +181,9 @@ static const OptionDesc ctrl_opt_desc[] = {
     }, {
         .name = "gid",
         .type = OPT_TYPE_GID_T,
+    }, {
+        .name = "terminate",
+        .type = OPT_TYPE_BOOLEAN,
     },
     END_OPTION_DESC
 };
@@ -226,8 +227,7 @@ static const OptionDesc locality_opt_desc[] = {
     {
         .name = "reject-locality-4",
         .type = OPT_TYPE_BOOLEAN,
-    },
-    {
+    }, {
         .name = "allow-set-locality",
         .type = OPT_TYPE_BOOLEAN,
     },
@@ -238,22 +238,21 @@ static const OptionDesc flags_opt_desc[] = {
     {
         .name = "not-need-init",
         .type = OPT_TYPE_BOOLEAN,
-    },
-    {
+    }, {
         .name = "startup-none",
         .type = OPT_TYPE_BOOLEAN,
-    },
-    {
+    }, {
         .name = "startup-clear",
         .type = OPT_TYPE_BOOLEAN,
-    },
-    {
+    }, {
         .name = "startup-state",
         .type = OPT_TYPE_BOOLEAN,
-    },
-    {
+    }, {
         .name = "startup-deactivated",
         .type = OPT_TYPE_BOOLEAN,
+    }, {
+        .name = "disable-auto-shutdown",
+        .type = OPT_TYPE_BOOLEAN,
     },
     END_OPTION_DESC
 };
@@ -268,6 +267,17 @@ static const OptionDesc seccomp_opt_desc[] = {
 };
 #endif
 
+static const OptionDesc migration_opt_desc[] = {
+    {
+        .name = "incoming",
+        .type = OPT_TYPE_BOOLEAN,
+    }, {
+        .name = "release-lock-outgoing",
+        .type = OPT_TYPE_BOOLEAN,
+    },
+    END_OPTION_DESC
+};
+
 /*
  * handle_log_options:
  * Parse and act upon the parsed log options. Initialize the logging.
@@ -623,13 +633,14 @@ handle_pid_options(char *options)
  * @options: the 'pid' options to parse
  * @tpmstatedir: Point to pointer for tpmstatedir
  * @mode: the mode of the TPM's state files
+ * @mode_is_default: true if user did not provide mode bits but using default
  * @tpmbackend_uri: Point to pointer for backend URI
  *
  * Returns 0 on success, -1 on failure.
  */
 static int
 parse_tpmstate_options(char *options, char **tpmstatedir, mode_t *mode,
-                       char **tpmbackend_uri)
+                       bool *mode_is_default, char **tpmbackend_uri)
 {
     OptionValues *ovs = NULL;
     char *error = NULL;
@@ -644,9 +655,14 @@ parse_tpmstate_options(char *options, char **tpmstatedir, mode_t *mode,
     }
 
     directory = option_get_string(ovs, "dir", NULL);
-    *mode = option_get_mode_t(ovs, "mode", 0640);
     backend_uri = option_get_string(ovs, "backend-uri", NULL);
 
+    /* Did user provide mode bits? User can only provide <= 0777 */
+    *mode = option_get_mode_t(ovs, "mode", 01000);
+    *mode_is_default = (*mode == 01000);
+    if (*mode_is_default)
+        *mode = 0640;
+
     if (directory) {
         *tpmstatedir = strdup(directory);
         if (!*tpmstatedir) {
@@ -693,12 +709,13 @@ handle_tpmstate_options(char *options)
     char *temp_uri = NULL;
     int ret = 0;
     mode_t mode;
+    bool mode_is_default = true;
 
     if (!options)
         return 0;
 
     if (parse_tpmstate_options(options, &tpmstatedir, &mode,
-                               &tpmbackend_uri) < 0) {
+                               &mode_is_default, &tpmbackend_uri) < 0) {
         ret = -1;
         goto error;
     }
@@ -714,7 +731,7 @@ handle_tpmstate_options(char *options)
         }
 
         if (tpmstate_set_backend_uri(temp_uri) < 0 ||
-            tpmstate_set_mode(mode) < 0) {
+            tpmstate_set_mode(mode, mode_is_default) < 0) {
             ret = -1;
             goto error;
         }
@@ -725,7 +742,7 @@ handle_tpmstate_options(char *options)
         }
         if ((strncmp(tpmbackend_uri, "dir://", 6) == 0 ||
             strncmp(tpmbackend_uri, "file://", 7)) &&
-            tpmstate_set_mode(mode) < 0) {
+            tpmstate_set_mode(mode, mode_is_default) < 0) {
             ret = -1;
             goto error;
         }
@@ -784,7 +801,7 @@ static int unixio_open_socket(const char *path, mode_t perm,
 
     if (chmod(su.sun_path, perm) < 0) {
         logprintf(STDERR_FILENO,
-                  "Could not change permssions on UnixIO socket: %s\n",
+                  "Could not change permissions on UnixIO socket: %s\n",
                   strerror(errno));
         goto error;
     }
@@ -918,10 +935,13 @@ error:
  * Parse the 'ctrl' (control channel) options.
  *
  * @options: the control channel options to parse
+ * @cc: pointer for pointer to allocated ctrlchannel struct
+ * @mainloop_flags: pointer to mainloop flags to add a flag to
  *
  * Returns 0 on success, -1 on failure.
  */
-static int parse_ctrlchannel_options(char *options, struct ctrlchannel **cc)
+static int parse_ctrlchannel_options(char *options, struct ctrlchannel **cc,
+                                     uint32_t *mainloop_flags)
 {
     OptionValues *ovs = NULL;
     char *error = NULL;
@@ -1014,13 +1034,16 @@ static int parse_ctrlchannel_options(char *options, struct ctrlchannel **cc)
             goto error;
         }
     } else {
-        logprintf(STDERR_FILENO, "Unsupport control channel type: %s\n", type);
+        logprintf(STDERR_FILENO, "Unsupported control channel type: %s\n", type);
         goto error;
     }
 
     if (*cc == NULL)
         goto error;
 
+    if (option_get_bool(ovs, "terminate", false))
+        *mainloop_flags |= MAIN_LOOP_FLAG_CTRL_END_ON_HUP;
+
     option_values_free(ovs);
 
     return 0;
@@ -1037,15 +1060,18 @@ error:
  * Parse and act upon the parsed 'ctrl' (control channel) options.
  *
  * @options: the control channel options to parse
+ * @cc: pointer for pointer to allocated ctrlchannel struct
+ * @mainloop_flags: pointer to mainloop flags to add a flag to
  *
  * Returns 0 on success, -1 on failure.
  */
-int handle_ctrlchannel_options(char *options, struct ctrlchannel **cc)
+int handle_ctrlchannel_options(char *options, struct ctrlchannel **cc,
+                               uint32_t *mainloop_flag)
 {
     if (!options)
         return 0;
 
-    if (parse_ctrlchannel_options(options, cc) < 0)
+    if (parse_ctrlchannel_options(options, cc, mainloop_flag) < 0)
         return -1;
 
     return 0;
@@ -1152,7 +1178,7 @@ static int parse_server_options(char *options, struct server **c)
             *c = server_new(fd, flags, NULL);
         }
     } else {
-        logprintf(STDERR_FILENO, "Unsupport socket type: %s\n", type);
+        logprintf(STDERR_FILENO, "Unsupported socket type: %s\n", type);
         goto error;
     }
 
@@ -1242,7 +1268,7 @@ int handle_locality_options(char *options, uint32_t *flags)
 }
 
 static int parse_flags_options(char *options, bool *need_init_cmd,
-                               uint16_t *startupType)
+                               uint16_t *startupType, bool *disable_auto_shutdown)
 {
     OptionValues *ovs = NULL;
     char *error = NULL;
@@ -1255,6 +1281,8 @@ static int parse_flags_options(char *options, bool *need_init_cmd,
 
     if (option_get_bool(ovs, "not-need-init", false))
         *need_init_cmd = false;
+    if (option_get_bool(ovs, "disable-auto-shutdown", false))
+        *disable_auto_shutdown = true;
 
     if (option_get_bool(ovs, "startup-clear", false))
         *startupType = TPM_ST_CLEAR;
@@ -1288,12 +1316,13 @@ error:
  * Returns 0 on success, -1 on failure.
  */
 int handle_flags_options(char *options, bool *need_init_cmd,
-                         uint16_t *startupType)
+                         uint16_t *startupType, bool *disable_auto_shutdown)
 {
     if (!options)
         return 0;
 
-    if (parse_flags_options(options, need_init_cmd, startupType) < 0)
+    if (parse_flags_options(options, need_init_cmd, startupType,
+                            disable_auto_shutdown) < 0)
         return -1;
 
     return 0;
@@ -1360,3 +1389,55 @@ int handle_seccomp_options(char *options, unsigned int *seccomp_action)
     return 0;
 }
 #endif /* WITH_SECCOMP */
+
+static int parse_migration_options(char *options, bool *incoming_migration,
+                                   bool *release_lock_outgoing)
+{
+    OptionValues *ovs = NULL;
+    char *error = NULL;
+
+    ovs = options_parse(options, migration_opt_desc, &error);
+    if (!ovs) {
+        logprintf(STDERR_FILENO, "Error parsing migration options: %s\n", error);
+        goto error;
+    }
+
+    *incoming_migration = option_get_bool(ovs, "incoming", false);
+    *release_lock_outgoing = option_get_bool(ovs, "release-lock-outgoing",
+                                             false);
+
+    option_values_free(ovs);
+
+    return 0;
+
+error:
+    option_values_free(ovs);
+    free(error);
+
+    return -1;
+}
+
+/*
+ * handle_migration_options:
+ * Parse the 'migration' options.
+ *
+ * @options: the migration options to parse
+ * @incoming_migration: whether there's an incoming migration
+ * @release_lock_outgoing: whether to release NVRAM lock on outgoing migration
+ *
+ * Return 0 on success, -1 on failure.
+ */
+int handle_migration_options(char *options, bool *incoming_migration,
+                             bool *release_lock_outgoing)
+{
+    *incoming_migration = false;
+
+    if (!options)
+        return 0;
+
+    if (parse_migration_options(options, incoming_migration,
+                                release_lock_outgoing) < 0)
+        return -1;
+
+    return 0;
+}
