@@ -128,10 +128,11 @@ int ctrlchannel_set_client_fd(struct ctrlchannel *cc, int fd)
     return clientfd;
 }
 
-static int ctrlchannel_return_state(ptm_getstate *pgs, int fd)
+static int ctrlchannel_return_state(ptm_getstate *pgs, int fd,
+                                    struct mainLoopParams *mlp)
 {
     uint32_t blobtype = be32toh(pgs->u.req.type);
-    const char *blobname = tpmlib_get_blobname(blobtype);
+    const char *blobname;
     uint32_t tpm_number = 0;
     unsigned char *blob = NULL;
     uint32_t blob_length = 0, return_length;
@@ -145,7 +146,11 @@ static int ctrlchannel_return_state(ptm_getstate *pgs, int fd)
     struct iovec iov[2];
     int iovcnt, n;
 
-    if (blobtype == PTM_BLOB_TYPE_VOLATILE)
+    blobname = tpmlib_get_blobname(blobtype);
+    if (!blobname)
+        res = TPM_FAIL;
+
+    if (res == 0 && blobtype == PTM_BLOB_TYPE_VOLATILE)
         res = SWTPM_NVRAM_Store_Volatile();
 
     if (res == 0)
@@ -194,6 +199,9 @@ static int ctrlchannel_return_state(ptm_getstate *pgs, int fd)
 
     free(blob);
 
+    if (fd >= 0 && blobtype == PTM_BLOB_TYPE_SAVESTATE)
+        mainloop_unlock_nvram(mlp, DEFAULT_LOCKING_RETRIES);
+
     return fd;
 }
 
@@ -230,7 +238,6 @@ static int ctrlchannel_receive_state(ptm_setstate *pss, ssize_t n, int fd)
         if (remain) {
             n = read_eintr(fd, pss->u.req.data, sizeof(pss->u.req.data));
             if (n < 0) {
-                res = TPM_IOERROR;
                 close(fd);
                 fd = -1;
                 goto err_fd_broken;
@@ -294,12 +301,12 @@ struct input {
  * ctrlchannel_recv_cmd: Receive a command on the control channel
  *
  * @fd: file descriptor for control channel
- * @msg: prepared msghdr struct for receiveing data with single
+ * @msg: prepared msghdr struct for receiving data with single
  *       msg_iov.
  *
  * This function returns 0 or a negative number if an error receiving
  * the command occurred, including a timeout. In case of success,
- * the nunber of bytes received is returned.
+ * the number of bytes received is returned.
  */
 static ssize_t ctrlchannel_recv_cmd(int fd,
                                     struct msghdr *msg)
@@ -449,7 +456,8 @@ static uint64_t get_ptm_caps_supported(TPMLIB_TPMVersion tpmversion)
             | PTM_CAP_SET_DATAFD
 #endif
             | PTM_CAP_SET_BUFFERSIZE
-            | PTM_CAP_GET_INFO;
+            | PTM_CAP_GET_INFO
+            | PTM_CAP_LOCK_STORAGE;
     if (tpmversion == TPMLIB_TPM_VERSION_2)
         caps |= PTM_CAP_SEND_COMMAND_HEADER;
 
@@ -468,9 +476,8 @@ static uint64_t get_ptm_caps_supported(TPMLIB_TPMVersion tpmversion)
  *            number when set via CMD_SET_LOCALITY
  * @tpm_running: indicates whether the TPM is running; may be changed by
  *               this function in case TPM is stopped or started
- * @mlp: mainloop parameters used; may be altered by this function incase of
+ * @mlp: mainloop parameters used; may be altered by this function in case of
  *       CMD_SET_DATAFD
- * @tpmversion: the emulated TPM's version
  *
  * This function returns the passed file descriptor or -1 in case the
  * file descriptor was closed.
@@ -514,6 +521,7 @@ int ctrlchannel_process_fd(int fd,
     ptm_loc *pl;
     ptm_setbuffersize *psbs;
     ptm_getinfo *pgi, _pgi;
+    ptm_lockstorage *pls;
 
     size_t out_len = 0;
     TPM_RESULT res;
@@ -552,13 +560,20 @@ int ctrlchannel_process_fd(int fd,
         if (n != (ssize_t)sizeof(ptm_init)) /* r/w */
             goto err_bad_input;
 
+        if (*tpm_running && !mlp->disable_auto_shutdown)
+            tpmlib_maybe_send_tpm2_shutdown(mlp->tpmversion,
+                                            &mlp->lastCommand);
+
         init_p = (ptm_init *)input.body;
 
         TPMLIB_Terminate();
 
         *tpm_running = false;
+
+        mlp->storage_locked = !mlp->incoming_migration;
+
         res = tpmlib_start(be32toh(init_p->u.req.init_flags),
-                           mlp->tpmversion);
+                           mlp->tpmversion, mlp->storage_locked);
         if (res) {
             logprintf(STDERR_FILENO,
                       "Error: Could not initialize the TPM\n");
@@ -574,6 +589,10 @@ int ctrlchannel_process_fd(int fd,
         if (n != 0) /* wo */
             goto err_bad_input;
 
+        if (*tpm_running && !mlp->disable_auto_shutdown)
+            tpmlib_maybe_send_tpm2_shutdown(mlp->tpmversion,
+                                            &mlp->lastCommand);
+
         TPMLIB_Terminate();
 
         *tpm_running = false;
@@ -586,8 +605,14 @@ int ctrlchannel_process_fd(int fd,
         if (n != 0) /* wo */
             goto err_bad_input;
 
+        if (*tpm_running && !mlp->disable_auto_shutdown)
+            tpmlib_maybe_send_tpm2_shutdown(mlp->tpmversion,
+                                            &mlp->lastCommand);
+
         TPMLIB_Terminate();
 
+        *tpm_running = false;
+
         *res_p = htobe32(TPM_SUCCESS);
         out_len = sizeof(ptm_res);
 
@@ -743,7 +768,7 @@ int ctrlchannel_process_fd(int fd,
         if (n < (ssize_t)sizeof(pgs->u.req)) /* rw */
             goto err_bad_input;
 
-        return ctrlchannel_return_state(pgs, fd);
+        return ctrlchannel_return_state(pgs, fd, mlp);
 
     case CMD_SET_STATEBLOB:
         if (*tpm_running)
@@ -751,6 +776,8 @@ int ctrlchannel_process_fd(int fd,
 
         /* tpm state dir must be set */
         SWTPM_NVRAM_Init();
+        if ((*terminate = !mainloop_ensure_locked_storage(mlp)))
+            goto err_io;
 
         pss = (ptm_setstate *)input.body;
         if (n < (ssize_t)offsetof(ptm_setstate, u.req.data)) /* rw */
@@ -858,6 +885,24 @@ int ctrlchannel_process_fd(int fd,
 
         break;
 
+    case CMD_LOCK_STORAGE:
+        if (n < (ssize_t)sizeof(pls->u.req)) /* rw */
+            goto err_bad_input;
+
+        pls = (ptm_lockstorage *)input.body;
+
+        mlp->locking_retries = be32toh(pls->u.req.retries);
+
+        pls = (ptm_lockstorage *)&output.body;
+        out_len = sizeof(pls->u.resp);
+
+        if (!mainloop_ensure_locked_storage(mlp))
+            pls->u.resp.tpm_result = htobe32(TPM_FAIL);
+        else
+            pls->u.resp.tpm_result = htobe32(TPM_SUCCESS);
+
+        break;
+
     default:
         logprintf(STDERR_FILENO,
                   "Error: Unknown command: 0x%08x\n", be32toh(input.cmd));
