@@ -70,7 +70,7 @@
 #include "locality.h"
 #include "logging.h"
 #include "tpm_ioctl.h"
-#include "swtpm_nvfile.h"
+#include "swtpm_nvstore.h"
 #include "tpmlib.h"
 #include "main.h"
 #include "utils.h"
@@ -78,6 +78,7 @@
 #include "seccomp_profile.h"
 #include "options.h"
 #include "capabilities.h"
+#include "swtpm_utils.h"
 
 /* maximum size of request buffer */
 #define TPM_REQ_MAX 4096
@@ -88,7 +89,7 @@ static TPMLIB_TPMVersion tpmversion;
 /* buffer containing the TPM request */
 static unsigned char *ptm_request;
 
-/* buffer containing the TPM response */
+/* buffer containing the TPM response; protected by file_ops_lock and thread_busy_lock */
 static unsigned char *ptm_response;
 
 /* offset from where to read from; reset when ptm_response is set */
@@ -100,7 +101,7 @@ static uint32_t ptm_req_len, ptm_res_len, ptm_res_tot;
 /* locality applied to TPM commands */
 static TPM_MODIFIER_INDICATOR locality;
 
-/* whether the TPM is running (TPM_Init was received) */
+/* whether the TPM is running (TPM_Init was received); protected by file_ops_lock */
 static bool tpm_running;
 
 /* flags on how to handle locality */
@@ -109,6 +110,26 @@ static uint32_t locality_flags;
 /* the fuse_session that we will signal an exit to to exit the prg. */
 static struct fuse_session *ptm_fuse_session;
 
+/* last command sent to the TPM */
+static uint32_t g_lastCommand = TPM_ORDINAL_NONE;
+
+/* TPM2_Shutdown() will NOT be sent by swtpm */
+static bool g_disable_auto_shutdown = false;
+
+/* whether to defer locking NVRAM storage due to incoming migration */
+static bool g_incoming_migration;
+
+/* whether NVRAM storage is currently locked */
+static bool g_storage_locked;
+
+/* whether to release the lock on outgoing migration */
+static bool g_release_lock_outgoing;
+
+/* how many times to retry locking; use for fallback after releasing
+   the lock on outgoing migration. */
+static unsigned int g_locking_retries;
+#define DEFAULT_LOCKING_RETRIES  300 /* 300 * 10ms */
+
 #if GLIB_MAJOR_VERSION >= 2
 # if GLIB_MINOR_VERSION >= 32
 
@@ -123,12 +144,13 @@ static GMutex *file_ops_lock;
 # endif
 #else
 
-#error Unsupport glib version
+#error Unsupported glib version
 
 #endif
 
 struct cuse_param {
     char *runas;
+    char *chroot;
     char *logging;
     char *keydata;
     char *migkeydata;
@@ -136,6 +158,7 @@ struct cuse_param {
     char *tpmstatedata;
     char *localitydata;
     char *seccompdata;
+    char *migrationdata;
     unsigned int seccomp_action;
     char *flagsdata;
     uint16_t startupType;
@@ -160,6 +183,7 @@ typedef struct stateblob_desc {
 } stateblob_desc;
 
 typedef enum tx_state_type {
+    TX_STATE_CLOSED = 0,
     TX_STATE_RW_COMMAND = 1,
     TX_STATE_SET_STATE_BLOB = 2,
     TX_STATE_GET_STATE_BLOB = 3,
@@ -218,18 +242,21 @@ static const char *usage =
 "                       the log file can be reset (truncate)\n"
 "--pid file=<path>|fd=<filedescriptor>\n"
 "                    :  write the process ID into the given file\n"
-"--tpmstate dir=<dir>[,mode=0...]\n"
-"                    :  set the directory where the TPM's state will be written\n"
+"--tpmstate dir=<dir>[,mode=0...]|backend-uri=<uri>\n"
+"                    :  set the directory or uri where the TPM's state will be written\n"
 "                       into; the TPM_PATH environment variable can be used\n"
-"                       instead;\n"
+"                       instead of dir option;\n"
 "                       mode allows a user to set the file mode bits of the state\n"
 "                       files; the default mode is 0640;\n"
-"--flags [not-need-init][,startup-clear|startup-state|startup-deactivated|startup-none]\n"
+"--flags [not-need-init][,startup-clear|startup-state|startup-deactivated|startup-none][,disable-auto-shutdown]\n"
 "                    :  not-need-init: commands can be sent without needing to\n"
 "                       send an INIT via control channel;\n"
 "                       startup-...: send Startup command with this type;\n"
+"                       disable-auto-shutdown disables automatic sending of\n"
+"                       TPM2_Shutdown before TPM 2 reset or swtpm termination;\n"
 "-r|--runas <user>   :  after creating the CUSE device, change to the given\n"
 "                       user\n"
+"-R|--chroot <path>  :  chroot to the given directory at startup\n"
 "--tpm2              :  choose TPM2 functionality\n"
 #ifdef WITH_SECCOMP
 # ifndef SCMP_ACT_LOG
@@ -240,12 +267,18 @@ static const char *usage =
 "                    :  Choose the action of the seccomp profile when a\n"
 "                       blacklisted syscall is executed; default is kill\n"
 #endif
-"--print-capabilites : print capabilities and terminate\n"
+"--migration [incoming][,release-lock-outgoing]\n"
+"                    : Incoming migration defers locking of storage backend\n"
+"                      until the TPM state is received; release-lock-outgoing\n"
+"                      releases the storage lock on outgoing migration\n"
+"--print-capabilities : print capabilities and terminate\n"
+"--print-states      : print existing TPM states and terminate\n"
 "-h|--help           :  display this help screen and terminate\n"
 "\n";
 
 static TPM_RESULT
-ptm_io_getlocality(TPM_MODIFIER_INDICATOR *loc, uint32_t tpmnum)
+ptm_io_getlocality(TPM_MODIFIER_INDICATOR *loc,
+                   uint32_t tpmnum SWTPM_ATTR_UNUSED)
 {
     *loc = locality;
     return TPM_SUCCESS;
@@ -266,6 +299,38 @@ static transfer_state tx_state;
 /* function prototypes */
 static void ptm_cleanup(void);
 
+/************************* NVRAM locking ************************************/
+
+/* ensure that the NVRAM is locked; returns false in case of failure */
+static bool ensure_locked_storage(void)
+{
+    TPM_RESULT res;
+
+    if (g_storage_locked)
+        return true;
+
+    /* if NVRAM hasn't been initialized yet locking may need to be retried */
+    res = SWTPM_NVRAM_Lock_Storage(g_locking_retries);
+    if (res == TPM_RETRY)
+        return true;
+    if (res != TPM_SUCCESS)
+        return false;
+
+    g_storage_locked = true;
+    g_incoming_migration = false;
+    g_locking_retries = 0;
+
+    return true;
+}
+
+static void unlock_nvram(unsigned int locking_retries)
+{
+    SWTPM_NVRAM_Unlock();
+
+    g_storage_locked = false;
+    g_locking_retries = locking_retries;
+}
+
 /************************* cached stateblob *********************************/
 
 static stateblob_desc cached_stateblob;
@@ -355,6 +420,9 @@ static TPM_RESULT cached_stateblob_load(uint32_t blobtype, TPM_BOOL decrypt)
         cached_stateblob.decrypt = decrypt;
     }
 
+    if (blobtype == PTM_BLOB_TYPE_SAVESTATE)
+        unlock_nvram(DEFAULT_LOCKING_RETRIES);
+
     return res;
 }
 
@@ -396,10 +464,12 @@ static int cached_stateblob_copy(void *dest, size_t destlen,
 /*
  * worker_thread: the worker thread
  */
-static void worker_thread(gpointer data, gpointer user_data)
+static void worker_thread(gpointer data, gpointer user_data SWTPM_ATTR_UNUSED)
 {
     struct thread_message *msg = (struct thread_message *)data;
 
+    /* file_ops_lock not needed since thread_busy is set */
+
     switch (msg->type) {
     case MESSAGE_TPM_CMD:
         TPMLIB_Process(&ptm_response, &ptm_res_len, &ptm_res_tot,
@@ -420,7 +490,7 @@ static void worker_thread(gpointer data, gpointer user_data)
  * tpm_start: Start the TPM
  *
  * Check whether the TPM's state directory exists and if it does
- * not exists, try to creat it. Start the thread pool, initilize
+ * not exists, try to create it. Start the thread pool, initialize
  * libtpms and allocate a global TPM request buffer.
  *
  * @flags: libtpms init flags
@@ -431,7 +501,8 @@ static int tpm_start(uint32_t flags, TPMLIB_TPMVersion l_tpmversion,
                      TPM_RESULT *res)
 {
     DIR *dir;
-    const char *tpmdir = tpmstate_get_dir();
+    const char *uri = tpmstate_get_backend_uri();
+    const char *tpmdir = uri + strlen("dir://");
 
     *res = TPM_FAIL;
 
@@ -466,7 +537,9 @@ static int tpm_start(uint32_t flags, TPMLIB_TPMVersion l_tpmversion,
         goto error_del_pool;
     }
 
-    *res = tpmlib_start(flags, l_tpmversion);
+    g_storage_locked = !g_incoming_migration;
+
+    *res = tpmlib_start(flags, l_tpmversion, g_storage_locked);
     if (*res != TPM_SUCCESS)
         goto error_del_pool;
 
@@ -499,7 +572,8 @@ static void ptm_write_fatal_error_response(TPMLIB_TPMVersion l_tpmversion)
 /*
  * ptm_send_startup: Send a TPM/TPM2_Startup
  */
-static int ptm_send_startup(uint16_t startupType, TPMLIB_TPMVersion l_tpmversion)
+static int ptm_send_startup(uint16_t startupType,
+                            TPMLIB_TPMVersion l_tpmversion SWTPM_ATTR_UNUSED)
 {
     uint32_t command_length;
     unsigned char command[sizeof(struct tpm_startup)];
@@ -512,6 +586,8 @@ static int ptm_send_startup(uint16_t startupType, TPMLIB_TPMVersion l_tpmversion
                                tpmversion,
                                command, max_command_length);
     if (command_length > 0) {
+        g_lastCommand = tpmlib_get_cmd_ordinal(command, command_length);
+
         rc = TPMLIB_Process(&ptm_response, &ptm_res_len, &ptm_res_tot,
                            (unsigned char *)command, command_length);
         ptm_read_offset = 0;
@@ -539,6 +615,9 @@ static void ptm_read_result(fuse_req_t req, size_t size)
 {
     size_t len = 0;
 
+    /* prevent other threads from reading or writing cmds or doing ioctls */
+    g_mutex_lock(FILE_OPS_LOCK);
+
     if (tpm_running) {
         /* wait until results are ready */
         worker_thread_wait_done();
@@ -547,12 +626,14 @@ static void ptm_read_result(fuse_req_t req, size_t size)
     if (ptm_read_offset < ptm_res_len) {
         len = ptm_res_len - ptm_read_offset;
         if (size < len)
-           len = size;
+            len = size;
     }
 
     fuse_reply_buf(req, (const char *)&ptm_response[ptm_read_offset], len);
 
     ptm_read_offset += len;
+
+    g_mutex_unlock(FILE_OPS_LOCK);
 }
 
 /*
@@ -597,8 +678,8 @@ static void ptm_read_stateblob(fuse_req_t req, size_t size)
  * Depending on the current state of the transfer interface (read/write)
  * return either the results of TPM commands or a data of a TPM state blob.
  */
-static void ptm_read(fuse_req_t req, size_t size, off_t off,
-                     struct fuse_file_info *fi)
+static void ptm_read(fuse_req_t req, size_t size, off_t off SWTPM_ATTR_UNUSED,
+                     struct fuse_file_info *fi SWTPM_ATTR_UNUSED)
 {
     switch (tx_state.state) {
     case TX_STATE_RW_COMMAND:
@@ -611,6 +692,9 @@ static void ptm_read(fuse_req_t req, size_t size, off_t off,
     case TX_STATE_GET_STATE_BLOB:
         ptm_read_stateblob(req, size);
         break;
+    case TX_STATE_CLOSED:
+        /* not possible */
+        break;
     }
 }
 
@@ -857,13 +941,21 @@ static void ptm_write_stateblob(fuse_req_t req, const char *buf, size_t size)
 static void ptm_write_cmd(fuse_req_t req, const char *buf, size_t size,
                           TPMLIB_TPMVersion l_tpmversion)
 {
+    uint32_t lastCommand;
+
     ptm_req_len = size;
     ptm_res_len = 0;
 
-    /* prevent other threads from writing or doing ioctls */
+    /* prevent other threads from reading or writing cmds or doing ioctls */
     g_mutex_lock(FILE_OPS_LOCK);
 
     if (tpm_running) {
+        /* before processing a command ensure that the NVRAM is locked */
+        if (!ensure_locked_storage()) {
+            fuse_reply_err(req, EIO);
+            goto cleanup;
+        }
+
         /* ensure that we only ever work on one TPM command */
         if (worker_thread_is_busy()) {
             fuse_reply_err(req, EBUSY);
@@ -882,6 +974,10 @@ static void ptm_write_cmd(fuse_req_t req, const char *buf, size_t size,
             goto skip_process;
         }
 
+        lastCommand = tpmlib_get_cmd_ordinal((unsigned char *)buf, ptm_req_len);
+        if (lastCommand != TPM_ORDINAL_NONE)
+            g_lastCommand = lastCommand;
+
         if (tpmlib_is_request_cancelable(l_tpmversion,
                                          (const unsigned char*)buf,
                                          ptm_req_len)) {
@@ -914,11 +1010,12 @@ cleanup:
 }
 
 /*
- * ptm_write: low-level write() interface; calls approriate function depending
+ * ptm_write: low-level write() interface; calls appropriate function depending
  *            on what is being transferred using the write()
  */
 static void ptm_write(fuse_req_t req, const char *buf, size_t size,
-                      off_t off, struct fuse_file_info *fi)
+                      off_t off SWTPM_ATTR_UNUSED,
+                      struct fuse_file_info *fi SWTPM_ATTR_UNUSED)
 {
     switch (tx_state.state) {
     case TX_STATE_RW_COMMAND:
@@ -931,6 +1028,9 @@ static void ptm_write(fuse_req_t req, const char *buf, size_t size,
     case TX_STATE_SET_STATE_BLOB:
         ptm_write_stateblob(req, buf, size);
         break;
+    case TX_STATE_CLOSED:
+        /* not possible */
+        break;
     }
 }
 
@@ -939,11 +1039,26 @@ static void ptm_write(fuse_req_t req, const char *buf, size_t size,
  */
 static void ptm_open(fuse_req_t req, struct fuse_file_info *fi)
 {
+    if (tx_state.state != TX_STATE_CLOSED) {
+        fuse_reply_err(req, EBUSY);
+        return;
+    }
+
     tx_state.state = TX_STATE_RW_COMMAND;
 
     fuse_reply_open(req, fi);
 }
 
+/*
+ * ptm_release:
+ */
+static void ptm_release(fuse_req_t req, struct fuse_file_info *fi)
+{
+    tx_state.state = TX_STATE_CLOSED;
+
+    fuse_reply_err(req, 0);
+}
+
 /*
  * ptm_ioctl : ioctl execution
  *
@@ -959,7 +1074,8 @@ static void ptm_open(fuse_req_t req, struct fuse_file_info *fi)
  *            needed buffer
  */
 static void ptm_ioctl(fuse_req_t req, int cmd, void *arg,
-                      struct fuse_file_info *fi, unsigned flags,
+                      struct fuse_file_info *fi SWTPM_ATTR_UNUSED,
+                      unsigned flags SWTPM_ATTR_UNUSED,
                       const void *in_buf, size_t in_bufsz, size_t out_bufsz)
 {
     TPM_RESULT res = TPM_FAIL;
@@ -967,6 +1083,9 @@ static void ptm_ioctl(fuse_req_t req, int cmd, void *arg,
     ptm_init *init_p;
     TPM_MODIFIER_INDICATOR orig_locality;
 
+    /* prevent other threads from writing or doing ioctls */
+    g_mutex_lock(FILE_OPS_LOCK);
+
     /* some commands have to wait until the worker thread is done */
     switch(cmd) {
     case PTM_GET_CAPABILITY:
@@ -974,6 +1093,7 @@ static void ptm_ioctl(fuse_req_t req, int cmd, void *arg,
     case PTM_CANCEL_TPM_CMD:
     case PTM_GET_CONFIG:
     case PTM_SET_BUFFERSIZE:
+    case PTM_LOCK_STORAGE:
         /* no need to wait */
         break;
     case PTM_INIT:
@@ -991,9 +1111,6 @@ static void ptm_ioctl(fuse_req_t req, int cmd, void *arg,
         break;
     }
 
-    /* prevent other threads from writing or doing ioctls */
-    g_mutex_lock(FILE_OPS_LOCK);
-
     switch (cmd) {
     case PTM_GET_CAPABILITY:
         if (out_bufsz != sizeof(ptm_cap)) {
@@ -1015,7 +1132,8 @@ static void ptm_ioctl(fuse_req_t req, int cmd, void *arg,
                     | PTM_CAP_STOP
                     | PTM_CAP_GET_CONFIG
                     | PTM_CAP_SET_BUFFERSIZE
-                    | PTM_CAP_GET_INFO;
+                    | PTM_CAP_GET_INFO
+                    | PTM_CAP_LOCK_STORAGE;
                 break;
             case TPMLIB_TPM_VERSION_1_2:
                 ptm_caps = PTM_CAP_INIT | PTM_CAP_SHUTDOWN
@@ -1030,7 +1148,8 @@ static void ptm_ioctl(fuse_req_t req, int cmd, void *arg,
                     | PTM_CAP_STOP
                     | PTM_CAP_GET_CONFIG
                     | PTM_CAP_SET_BUFFERSIZE
-                    | PTM_CAP_GET_INFO;
+                    | PTM_CAP_GET_INFO
+                    | PTM_CAP_LOCK_STORAGE;
                 break;
             }
             fuse_reply_ioctl(req, 0, &ptm_caps, sizeof(ptm_caps));
@@ -1046,6 +1165,10 @@ static void ptm_ioctl(fuse_req_t req, int cmd, void *arg,
 
             worker_thread_end();
 
+            if (tpm_running && !g_disable_auto_shutdown)
+                tpmlib_maybe_send_tpm2_shutdown(tpmversion,
+                                                &g_lastCommand);
+
             TPMLIB_Terminate();
 
             tpm_running = false;
@@ -1063,6 +1186,9 @@ static void ptm_ioctl(fuse_req_t req, int cmd, void *arg,
     case PTM_STOP:
         worker_thread_end();
 
+        if (tpm_running && !g_disable_auto_shutdown)
+            tpmlib_maybe_send_tpm2_shutdown(tpmversion, &g_lastCommand);
+
         res = TPM_SUCCESS;
         TPMLIB_Terminate();
 
@@ -1078,9 +1204,14 @@ static void ptm_ioctl(fuse_req_t req, int cmd, void *arg,
     case PTM_SHUTDOWN:
         worker_thread_end();
 
+        if (tpm_running && !g_disable_auto_shutdown)
+            tpmlib_maybe_send_tpm2_shutdown(tpmversion, &g_lastCommand);
+
         res = TPM_SUCCESS;
         TPMLIB_Terminate();
 
+        tpm_running = false;
+
         free(ptm_response);
         ptm_response = NULL;
 
@@ -1223,8 +1354,9 @@ static void ptm_ioctl(fuse_req_t req, int cmd, void *arg,
 
         /* tpm state dir must be set */
         SWTPM_NVRAM_Init();
-
-        if (in_bufsz != sizeof(ptm_setstate)) {
+        if ((exit_prg = !ensure_locked_storage())) {
+            fuse_reply_err(req, EIO);
+        } else if (in_bufsz != sizeof(ptm_setstate)) {
             struct iovec iov = { arg, sizeof(uint32_t) };
             fuse_reply_ioctl_retry(req, &iov, 1, NULL, 0);
         } else {
@@ -1308,6 +1440,25 @@ static void ptm_ioctl(fuse_req_t req, int cmd, void *arg,
         }
         break;
 
+    case PTM_LOCK_STORAGE:
+        if (out_bufsz != sizeof(ptm_lockstorage)) {
+            struct iovec iov = { arg, sizeof(uint32_t) };
+            fuse_reply_ioctl_retry(req, &iov, 1, NULL, 0);
+        } else {
+            ptm_lockstorage *in_pls = (ptm_lockstorage *)in_buf;
+            ptm_lockstorage out_pls;
+
+            g_locking_retries = in_pls->u.req.retries;
+
+            if (!ensure_locked_storage())
+                out_pls.u.resp.tpm_result = TPM_FAIL;
+            else
+                out_pls.u.resp.tpm_result = TPM_SUCCESS;
+            fuse_reply_ioctl(req, 0, &out_pls, sizeof(out_pls));
+        }
+
+        break;
+
     default:
         fuse_reply_err(req, EINVAL);
     }
@@ -1385,6 +1536,7 @@ static const struct cuse_lowlevel_ops clops = {
     .open = ptm_open,
     .read = ptm_read,
     .write = ptm_write,
+    .release = ptm_release,
     .ioctl = ptm_ioctl,
     .init_done = ptm_init_done,
 };
@@ -1433,6 +1585,7 @@ int swtpm_cuse_main(int argc, char **argv, const char *prgname, const char *ifac
         {"min"           , required_argument, 0, 'm'},
         {"name"          , required_argument, 0, 'n'},
         {"runas"         , required_argument, 0, 'r'},
+        {"chroot"        , required_argument, 0, 'R'},
         {"log"           , required_argument, 0, 'l'},
         {"locality"      , required_argument, 0, 'L'},
         {"key"           , required_argument, 0, 'k'},
@@ -1446,8 +1599,10 @@ int swtpm_cuse_main(int argc, char **argv, const char *prgname, const char *ifac
 #ifdef WITH_SECCOMP
         {"seccomp"       , required_argument, 0, 'S'},
 #endif
+        {"migration"     , required_argument, 0, 'i'},
         {"print-capabilities"
                          ,       no_argument, 0, 'a'},
+        {"print-states"  ,       no_argument, 0, 'e'},
         {NULL            , 0                , 0, 0  },
     };
     struct cuse_info cinfo;
@@ -1458,23 +1613,29 @@ int swtpm_cuse_main(int argc, char **argv, const char *prgname, const char *ifac
     char *cinfo_argv[1] = { 0 };
     unsigned int num;
     struct passwd *passwd;
-    const char *tpmdir;
+    const char *uri = NULL;
     int n, tpmfd;
     char path[PATH_MAX];
     int ret = 0;
     bool printcapabilities = false;
+    bool printstates = false;
     bool need_init_cmd = true;
     TPM_RESULT res;
 
     memset(&cinfo, 0, sizeof(cinfo));
     memset(&param, 0, sizeof(param));
+#if GLIB_MINOR_VERSION >= 32
+    g_mutex_init(FILE_OPS_LOCK);
+#else
+    FILE_OPS_LOCK = g_mutex_new();
+#endif
 
     log_set_prefix("swtpm: ");
 
     tpmversion = TPMLIB_TPM_VERSION_1_2;
 
     while (true) {
-        opt = getopt_long(argc, argv, "M:m:n:r:hv", longopts, &longindex);
+        opt = getopt_long(argc, argv, "M:m:n:r:R:hv", longopts, &longindex);
 
         if (opt == -1)
             break;
@@ -1528,6 +1689,9 @@ int swtpm_cuse_main(int argc, char **argv, const char *prgname, const char *ifac
         case 'r': /* runas */
             param.runas = optarg;
             break;
+        case 'R':
+            param.chroot = optarg;
+            break;
         case 'l': /* log */
             param.logging = optarg;
             break;
@@ -1555,12 +1719,18 @@ int swtpm_cuse_main(int argc, char **argv, const char *prgname, const char *ifac
         case 'S':
             param.seccompdata = optarg;
             break;
+        case 'i': /* --migration */
+            param.migrationdata = optarg;
+            break;
         case 'h': /* help */
             fprintf(stdout, usage, prgname, iface);
             goto exit;
         case 'a':
             printcapabilities = true;
             break;
+        case 'e':
+            printstates = true;
+            break;
         case 'v': /* version */
             fprintf(stdout, "TPM emulator CUSE interface version %d.%d.%d, "
                     "Copyright (c) 2014-2015 IBM Corp.\n",
@@ -1578,43 +1748,6 @@ int swtpm_cuse_main(int argc, char **argv, const char *prgname, const char *ifac
         goto exit;
     }
 
-    /*
-     * choose the TPM version early so that getting/setting
-     * buffer size works.
-     */
-    if (TPMLIB_ChooseTPMVersion(tpmversion) != TPM_SUCCESS) {
-        logprintf(STDERR_FILENO,
-                  "Error: Could not choose TPM version.\n");
-        ret = EXIT_FAILURE;
-        goto exit;
-    }
-
-    if (printcapabilities) {
-        ret = capabilities_print_json(true) ? EXIT_FAILURE : EXIT_SUCCESS;
-        goto exit;
-    }
-
-    SWTPM_NVRAM_Set_TPMVersion(tpmversion);
-
-    if (!cinfo.dev_info_argv) {
-        logprintf(STDERR_FILENO, "Error: device name missing\n");
-        ret = -2;
-        goto exit;
-    }
-
-    if (handle_log_options(param.logging) < 0 ||
-        handle_key_options(param.keydata) < 0 ||
-        handle_migration_key_options(param.migkeydata) < 0 ||
-        handle_pid_options(param.piddata) < 0 ||
-        handle_tpmstate_options(param.tpmstatedata) < 0 ||
-        handle_seccomp_options(param.seccompdata, &param.seccomp_action) < 0 ||
-        handle_locality_options(param.localitydata, &locality_flags) < 0 ||
-        handle_flags_options(param.flagsdata, &need_init_cmd,
-                             &param.startupType) < 0) {
-        ret = -3;
-        goto exit;
-    }
-
     if (setuid(0)) {
         logprintf(STDERR_FILENO, "Error: Unable to setuid root. uid = %d, "
                   "euid = %d, gid = %d\n", getuid(), geteuid(), getgid());
@@ -1622,6 +1755,13 @@ int swtpm_cuse_main(int argc, char **argv, const char *prgname, const char *ifac
         goto exit;
     }
 
+    if (param.chroot) {
+        if (do_chroot(param.chroot) < 0) {
+            ret = EXIT_FAILURE;
+            goto exit;
+        }
+    }
+
     if (param.runas) {
         if (!(passwd = getpwnam(param.runas))) {
             logprintf(STDERR_FILENO, "User '%s' does not exist\n",
@@ -1631,8 +1771,65 @@ int swtpm_cuse_main(int argc, char **argv, const char *prgname, const char *ifac
         }
     }
 
-    tpmdir = tpmstate_get_dir();
-    if (tpmdir == NULL) {
+    if (handle_log_options(param.logging) < 0) {
+        ret = EXIT_FAILURE;
+        goto exit;
+    }
+
+    if (printcapabilities) {
+        /*
+         * Choose the TPM version so that getting/setting buffer size works.
+         * Ignore failure, for backward compatibility when TPM 1.2 is disabled.
+         */
+        ret = capabilities_print_json(true, tpmversion) ? EXIT_FAILURE : EXIT_SUCCESS;
+        goto exit;
+    }
+
+    if (tpmlib_choose_tpm_version(tpmversion) != TPM_SUCCESS) {
+        ret = EXIT_FAILURE;
+        goto exit;
+    }
+
+    tpmstate_set_version(tpmversion);
+
+    if (printstates) {
+        if (handle_tpmstate_options(param.tpmstatedata) < 0) {
+            ret = EXIT_FAILURE;
+            goto exit;
+        }
+        if (param.tpmstatedata == NULL) {
+            logprintf(STDERR_FILENO,
+                      "Error: --tpmstate option is required for --print-states\n");
+            ret = EXIT_FAILURE;
+            goto exit;
+        }
+        ret = SWTPM_NVRAM_PrintJson();
+        ret = ret ? EXIT_FAILURE : EXIT_SUCCESS;
+        goto exit;
+    }
+
+    if (!cinfo.dev_info_argv) {
+        logprintf(STDERR_FILENO, "Error: device name missing\n");
+        ret = -2;
+        goto exit;
+    }
+
+    if (handle_key_options(param.keydata) < 0 ||
+        handle_migration_key_options(param.migkeydata) < 0 ||
+        handle_pid_options(param.piddata) < 0 ||
+        handle_tpmstate_options(param.tpmstatedata) < 0 ||
+        handle_seccomp_options(param.seccompdata, &param.seccomp_action) < 0 ||
+        handle_locality_options(param.localitydata, &locality_flags) < 0 ||
+        handle_flags_options(param.flagsdata, &need_init_cmd,
+                             &param.startupType, &g_disable_auto_shutdown) < 0 ||
+        handle_migration_options(param.migrationdata, &g_incoming_migration,
+                                 &g_release_lock_outgoing) < 0) {
+        ret = -3;
+        goto exit;
+    }
+
+    uri = tpmstate_get_backend_uri();
+    if (uri == NULL) {
         logprintf(STDERR_FILENO,
                   "Error: No TPM state directory is defined; "
                   "TPM_PATH is not set\n");
@@ -1671,10 +1868,12 @@ int swtpm_cuse_main(int argc, char **argv, const char *prgname, const char *ifac
 
     worker_thread_init();
 
+    g_mutex_lock(FILE_OPS_LOCK);
+
     if (!need_init_cmd) {
         if (tpm_start(0, tpmversion, &res) < 0) {
             ret = -1;
-            goto exit;
+            goto err_unlock;
         }
         tpm_running = true;
     }
@@ -1682,21 +1881,25 @@ int swtpm_cuse_main(int argc, char **argv, const char *prgname, const char *ifac
     if (param.startupType != _TPM_ST_NONE) {
         if (ptm_send_startup(param.startupType, tpmversion) < 0) {
             ret = -1;
-            goto exit;
+            goto err_unlock;
         }
     }
 
-#if GLIB_MINOR_VERSION >= 32
-    g_mutex_init(FILE_OPS_LOCK);
-#else
-    FILE_OPS_LOCK = g_mutex_new();
-#endif
+    g_mutex_unlock(FILE_OPS_LOCK);
 
     ret = ptm_cuse_lowlevel_main(1, argv, &cinfo, &clops, &param);
 
 exit:
     ptm_cleanup();
     free(cinfo_argv[0]);
+#if GLIB_MINOR_VERSION < 32
+    g_mutex_free(FILE_OPS_LOCK);
+#endif
 
     return ret;
+
+err_unlock:
+    g_mutex_unlock(FILE_OPS_LOCK);
+
+    goto exit;
 }
