@@ -66,6 +66,7 @@
 #include "sys_dependencies.h"
 #include "compiler_dependencies.h"
 #include "swtpm_utils.h"
+#include "swtpm_nvstore.h"
 
 /* local variables */
 static TPM_MODIFIER_INDICATOR locality;
@@ -82,6 +83,37 @@ mainloop_cb_get_locality(TPM_MODIFIER_INDICATOR *loc,
     return TPM_SUCCESS;
 }
 
+/* ensure that the storage is locked; returns false in case of failure */
+bool mainloop_ensure_locked_storage(struct mainLoopParams *mlp)
+{
+    TPM_RESULT res;
+
+    if (mlp->storage_locked)
+        return true;
+
+    /* if NVRAM hasn't been initialized yet locking may need to be retried */
+    res = SWTPM_NVRAM_Lock_Storage(mlp->locking_retries);
+    if (res == TPM_RETRY)
+        return true;
+    if (res != TPM_SUCCESS)
+        return false;
+
+    mlp->locking_retries = 0;
+    mlp->storage_locked = true;
+    mlp->incoming_migration = false;
+
+    return true;
+}
+
+void mainloop_unlock_nvram(struct mainLoopParams *mlp,
+                           unsigned int locking_retries)
+{
+    SWTPM_NVRAM_Unlock();
+
+    mlp->storage_locked = false;
+    mlp->locking_retries = locking_retries;
+}
+
 int mainLoop(struct mainLoopParams *mlp,
              int notify_fd)
 {
@@ -102,6 +134,7 @@ int mainLoop(struct mainLoopParams *mlp,
     struct iovec        iov[3];
     uint32_t            ack = htobe32(0);
     struct tpm2_resp_prefix respprefix;
+    uint32_t            lastCommand;
 
     /* poolfd[] indexes */
     enum {
@@ -141,9 +174,11 @@ int mainLoop(struct mainLoopParams *mlp,
                                   mlp->startupType,
                                   mlp->tpmversion,
                                   command, max_command_length);
-        if (command_length > 0)
+        if (command_length > 0) {
+            mlp->lastCommand = tpmlib_get_cmd_ordinal(command, command_length);
             rc = TPMLIB_Process(&rbuffer, &rlength, &rTotal,
                                 command, command_length);
+        }
 
         if (rc || command_length == 0) {
             mainloop_terminate = true;
@@ -225,6 +260,10 @@ int mainLoop(struct mainLoopParams *mlp,
                                                     &mainloop_terminate,
                                                     &locality, &tpm_running,
                                                     mlp);
+                if (ctrlclntfd < 0 &&
+                    mlp->flags & MAIN_LOOP_FLAG_CTRL_END_ON_HUP)
+                    mainloop_terminate = true;
+
                 if (mainloop_terminate)
                     break;
             }
@@ -233,11 +272,20 @@ int mainLoop(struct mainLoopParams *mlp,
                 if (ctrlclntfd >= 0)
                     close(ctrlclntfd);
                 ctrlclntfd = -1;
+                /* unixio gets this signal, not tcp */
+                if (mlp->flags & MAIN_LOOP_FLAG_CTRL_END_ON_HUP) {
+                    mainloop_terminate = true;
+                    break;
+                }
             }
 
             if (!(pollfds[DATA_CLIENT_FD].revents & POLLIN))
                 continue;
 
+            /* before processing a command ensure that the storage is locked */
+            if ((mainloop_terminate = !mainloop_ensure_locked_storage(mlp)))
+                break;
+
             /* Read the command.  The number of bytes is determined by 'paramSize' in the stream */
             if (rc == 0) {
                 rc = SWTPM_IO_Read(&connection_fd, command, &command_length,
@@ -273,6 +321,14 @@ int mainLoop(struct mainLoopParams *mlp,
                 }
             }
 
+            if (rc == 0) {
+                lastCommand =
+                    tpmlib_get_cmd_ordinal(&command[cmd_offset],
+                                           command_length - cmd_offset);
+                if (lastCommand != TPM_ORDINAL_NONE)
+                    mlp->lastCommand = lastCommand;
+            }
+
             if (rc == 0) {
                 rlength = 0;                                /* clear the response buffer */
                 rc = tpmlib_process(&rbuffer,
@@ -319,6 +375,9 @@ skip_process:
             break;
     }
 
+    if (tpm_running && !mlp->disable_auto_shutdown)
+        tpmlib_maybe_send_tpm2_shutdown(mlp->tpmversion, &mlp->lastCommand);
+
     free(rbuffer);
     free(command);
 
@@ -326,6 +385,9 @@ skip_process:
         close(ctrlclntfd);
     ctrlchannel_set_client_fd(mlp->cc, -1);
 
+    if (connection_fd.fd >= 0 && !(mlp->flags & MAIN_LOOP_FLAG_USE_FD))
+        close(connection_fd.fd);
+
     if (mlp->fd >= 0) {
         close(mlp->fd);
         mlp->fd = -1;
