@@ -57,7 +57,7 @@
 #include "main.h"
 #include "swtpm_debug.h"
 #include "swtpm_io.h"
-#include "swtpm_nvfile.h"
+#include "swtpm_nvstore.h"
 #include "server.h"
 #include "common.h"
 #include "logging.h"
@@ -68,7 +68,7 @@
 #include "ctrlchannel.h"
 #include "tpmstate.h"
 #include "sys_dependencies.h"
-#include "osx.h"
+#include "daemonize.h"
 #include "seccomp_profile.h"
 #include "options.h"
 #include "capabilities.h"
@@ -105,9 +105,10 @@ static void usage(FILE *file, const char *prgname, const char *iface)
     "\n"
     "-p|--port <port> : use the given port\n"
     "-f|--fd <fd>     : use the given socket file descriptor\n"
-    "-t|--terminate   : terminate the TPM once a connection has been lost\n"
+    "-t|--terminate   : terminate the TPM once the data channel connection (TCP)\n"
+    "                   has been lost\n"
     "-d|--daemon      : daemonize the TPM\n"
-    "--ctrl type=[unixio|tcp][,path=<path>][,port=<port>[,bindaddr=address[,ifname=ifname]]][,fd=<filedescriptor>|clientfd=<filedescriptor>][,mode=0...][,uid=uid][,gid=gid]\n"
+    "--ctrl type=[unixio|tcp][,path=<path>][,port=<port>[,bindaddr=address[,ifname=ifname]]][,fd=<filedescriptor>|clientfd=<filedescriptor>][,mode=0...][,uid=uid][,gid=gid][,terminate]\n"
     "                 : TPM control channel using either UnixIO or TCP sockets;\n"
     "                   the path is only valid for Unixio channels; the port must\n"
     "                   be given in case the type is TCP; the TCP socket is bound\n"
@@ -121,6 +122,7 @@ static void usage(FILE *file, const char *prgname, const char *iface)
     "                   mode allows a user to set the file mode bits of a Unixio socket;\n"
     "                   the value must be given in octal number format\n"
     "                   uid and gid set the ownership of the Unixio socket's file;\n"
+    "                   terminate terminates on ctrl channel connection loss;\n"
     "--migration-key file=<path>|fd=<fd>[,mode=aes-cbc|aes-256-cbc][,format=hex|binary][,remove=[true|false]]\n"
     "                 : use an AES key for the encryption of the TPM's state\n"
     "                   when it is retrieved from the TPM via ioctls;\n"
@@ -149,10 +151,10 @@ static void usage(FILE *file, const char *prgname, const char *iface)
     "                   allow-set-locality: accept SetLocality command\n"
     "--pid file=<path>|fd=<filedescriptor>\n"
     "                 : write the process ID into the given file\n"
-    "--tpmstate dir=<dir>[,mode=0...]\n"
-    "                 : set the directory where the TPM's state will be written\n"
+    "--tpmstate dir=<dir>[,mode=0...]|backend-uri=<uri>\n"
+    "                 : set the directory or uri where the TPM's state will be written\n"
     "                   into; the TPM_PATH environment variable can be used\n"
-    "                   instead;\n"
+    "                   instead dir option;\n"
     "                   mode allows a user to set the file mode bits of the state files;\n"
     "                   the default mode is 0640;\n"
     "--server [type=tcp][,port=port[,bindaddr=address[,ifname=ifname]]][,fd=fd][,disconnect]\n"
@@ -168,11 +170,15 @@ static void usage(FILE *file, const char *prgname, const char *iface)
     "                   mode allows a user to set the file mode bits of the socket; the\n"
     "                   value must be given in octal number format;\n"
     "                   uid and gid set the ownership of the Unixio socket's file;\n"
-    "--flags [not-need-init][,startup-clear|startup-state|startup-deactivated|startup-none]\n"
+    "--flags [not-need-init][,startup-clear|startup-state|startup-deactivated|startup-none][,disable-auto-shutdown]\n"
     "                 : not-need-init: commands can be sent without needing to\n"
     "                   send an INIT via control channel;\n"
     "                   startup-...: send Startup command with this type;\n"
+    "                   disable-auto-shutdown disables automatic sending of\n"
+    "                   TPM2_Shutdown before TPM 2 reset or swtpm termination;\n"
     "-r|--runas <user>: change to the given user\n"
+    "-R|--chroot <path>\n"
+    "                 : chroot to the given directory at startup\n"
     "--tpm2           : choose TPM2 functionality\n"
 #ifdef WITH_SECCOMP
 # ifndef SCMP_ACT_LOG
@@ -183,8 +189,14 @@ static void usage(FILE *file, const char *prgname, const char *iface)
     "                 : Choose the action of the seccomp profile when a\n"
     "                   blacklisted syscall is executed; default is kill\n"
 #endif
-    "--print-capabilites\n"
+    "--migration [incoming][,release-lock-outgoing]\n"
+    "                 : Incoming migration defers locking of storage backend\n"
+    "                   until the TPM state is received; release-lock-outgoing\n"
+    "                   releases the storage lock on outgoing migration\n"
+    "--print-capabilities\n"
     "                 : print capabilities and terminate\n"
+    "--print-states\n"
+    "                 : print existing TPM states and terminate\n"
     "-h|--help        : display this help screen and terminate\n"
     "\n",
     prgname, iface);
@@ -213,6 +225,10 @@ int swtpm_main(int argc, char **argv, const char *prgname, const char *iface)
         .locality_flags = 0,
         .tpmversion = TPMLIB_TPM_VERSION_1_2,
         .startupType = _TPM_ST_NONE,
+        .lastCommand = TPM_ORDINAL_NONE,
+        .disable_auto_shutdown = false,
+        .incoming_migration = false,
+        .storage_locked = false,
     };
     struct server *server = NULL;
     unsigned long val;
@@ -228,13 +244,16 @@ int swtpm_main(int argc, char **argv, const char *prgname, const char *iface)
     char *serverdata = NULL;
     char *flagsdata = NULL;
     char *seccompdata = NULL;
+    char *migrationdata = NULL;
     char *runas = NULL;
+    char *chroot = NULL;
     bool need_init_cmd = true;
 #ifdef DEBUG
     time_t              start_time;
 #endif
     unsigned int seccomp_action;
     bool printcapabilities = false;
+    bool printstates = false;
     static struct option longopts[] = {
         {"daemon"    ,       no_argument, 0, 'd'},
         {"help"      ,       no_argument, 0, 'h'},
@@ -242,6 +261,7 @@ int swtpm_main(int argc, char **argv, const char *prgname, const char *iface)
         {"fd"        , required_argument, 0, 'f'},
         {"server"    , required_argument, 0, 'c'},
         {"runas"     , required_argument, 0, 'r'},
+        {"chroot"    , required_argument, 0, 'R'},
         {"terminate" ,       no_argument, 0, 't'},
         {"locality"  , required_argument, 0, 'L'},
         {"log"       , required_argument, 0, 'l'},
@@ -255,15 +275,17 @@ int swtpm_main(int argc, char **argv, const char *prgname, const char *iface)
 #ifdef WITH_SECCOMP
         {"seccomp"   , required_argument, 0, 'S'},
 #endif
+        {"migration" , required_argument, 0, 'i'},
         {"print-capabilities"
                      ,       no_argument, 0, 'a'},
+        {"print-states",     no_argument, 0, 'e'},
         {NULL        , 0                , 0, 0  },
     };
 
     log_set_prefix("swtpm: ");
 
     while (TRUE) {
-        opt = getopt_long(argc, argv, "dhp:f:tr:", longopts, &longindex);
+        opt = getopt_long(argc, argv, "dhp:f:tr:R:", longopts, &longindex);
 
         if (opt == -1)
             break;
@@ -271,6 +293,11 @@ int swtpm_main(int argc, char **argv, const char *prgname, const char *iface)
         switch (opt) {
         case 'd':
             daemonize = TRUE;
+            if (daemonize_prep() == -1) {
+                logprintf(STDERR_FILENO,
+                          "Could not prepare to daemonize: %s\n", strerror(errno));
+                exit(EXIT_FAILURE);
+            }
             break;
 
         case 'p':
@@ -378,14 +405,26 @@ int swtpm_main(int argc, char **argv, const char *prgname, const char *iface)
             printcapabilities = true;
             break;
 
+        case 'e':
+            printstates = true;
+            break;
+
         case 'r':
             runas = optarg;
             break;
 
+        case 'R':
+            chroot = optarg;
+            break;
+
         case 'S':
             seccompdata = optarg;
             break;
 
+        case 'i':
+            migrationdata = optarg;
+            break;
+
         default:
             usage(stderr, prgname, iface);
             exit(EXIT_FAILURE);
@@ -398,6 +437,20 @@ int swtpm_main(int argc, char **argv, const char *prgname, const char *iface)
         exit(EXIT_FAILURE);
     }
 
+    if (chroot) {
+        if (do_chroot(chroot) < 0)
+            exit(EXIT_FAILURE);
+    }
+
+    /* change process ownership before accessing files */
+    if (runas) {
+        if (change_process_owner(runas) < 0)
+            exit(EXIT_FAILURE);
+    }
+
+    if (handle_log_options(logdata) < 0)
+        exit(EXIT_FAILURE);
+
     if (mlp.fd >= 0 && mlp.fd < 3) {
         /* no std{in,out,err} */
         logprintf(STDERR_FILENO,
@@ -405,43 +458,50 @@ int swtpm_main(int argc, char **argv, const char *prgname, const char *iface)
         exit(EXIT_FAILURE);
     }
 
-    /*
-     * choose the TPM version early so that getting/setting
-     * buffer size works.
-     */
-    if (TPMLIB_ChooseTPMVersion(mlp.tpmversion) != TPM_SUCCESS) {
-        logprintf(STDERR_FILENO,
-                  "Error: Could not choose TPM version.\n");
-        exit(EXIT_FAILURE);
-    }
-
     if (printcapabilities) {
-        ret = capabilities_print_json(false);
+        /*
+         * Choose the TPM version so that getting/setting buffer size works.
+         * Ignore failure, for backward compatibility when TPM 1.2 is disabled.
+         */
+        ret = capabilities_print_json(false, mlp.tpmversion);
         exit(ret ? EXIT_FAILURE : EXIT_SUCCESS);
     }
 
-    if (handle_ctrlchannel_options(ctrlchdata, &mlp.cc) < 0 ||
+    if (tpmlib_choose_tpm_version(mlp.tpmversion) != TPM_SUCCESS)
+        exit(EXIT_FAILURE);
+
+    if (handle_ctrlchannel_options(ctrlchdata, &mlp.cc, &mlp.flags) < 0 ||
         handle_server_options(serverdata, &server) < 0) {
         goto exit_failure;
     }
 
-    /* change process ownership before accessing files */
-    if (runas) {
-        if (change_process_owner(runas) < 0)
+    tpmstate_set_version(mlp.tpmversion);
+
+    if (printstates) {
+        if (handle_tpmstate_options(tpmstatedata) < 0)
+            goto exit_failure;
+        if (tpmstatedata == NULL) {
+            logprintf(STDERR_FILENO,
+                      "Error: --tpmstate option is required for --print-states\n");
+            goto exit_failure;
+        }
+        ret = SWTPM_NVRAM_PrintJson();
+        if (ret == 0)
+            goto exit_success;
+        else
             goto exit_failure;
     }
 
-    SWTPM_NVRAM_Set_TPMVersion(mlp.tpmversion);
-
-    if (handle_log_options(logdata) < 0 ||
-        handle_key_options(keydata) < 0 ||
+    if (handle_key_options(keydata) < 0 ||
         handle_migration_key_options(migkeydata) < 0 ||
         handle_pid_options(piddata) < 0 ||
         handle_locality_options(localitydata, &mlp.locality_flags) < 0 ||
         handle_tpmstate_options(tpmstatedata) < 0 ||
         handle_seccomp_options(seccompdata, &seccomp_action) < 0 ||
         handle_flags_options(flagsdata, &need_init_cmd,
-                             &mlp.startupType) < 0) {
+                             &mlp.startupType, &mlp.disable_auto_shutdown) < 0 ||
+        handle_migration_options(migrationdata, &mlp.incoming_migration,
+                                 &mlp.release_lock_outgoing) < 0) {
         goto exit_failure;
     }
 
@@ -459,17 +519,6 @@ int swtpm_main(int argc, char **argv, const char *prgname, const char *iface)
             mlp.flags |= MAIN_LOOP_FLAG_TERMINATE | MAIN_LOOP_FLAG_USE_FD;
     }
 
-    if (daemonize) {
-#ifdef __APPLE__
-        if (0 != osx_daemon(1, 0)) {
-#else
-        if (0 != daemon(1, 0)) {
-#endif
-            logprintf(STDERR_FILENO, "Error: Could not daemonize.\n");
-            goto exit_failure;
-        }
-    }
-
     if (pidfile_write(getpid()) < 0) {
         goto exit_failure;
     }
@@ -489,7 +538,9 @@ int swtpm_main(int argc, char **argv, const char *prgname, const char *iface)
         goto error_no_tpm;
 
     if (!need_init_cmd) {
-        if ((rc = tpmlib_start(0, mlp.tpmversion)))
+        mlp.storage_locked = !mlp.incoming_migration;
+
+        if ((rc = tpmlib_start(0, mlp.tpmversion, mlp.storage_locked)))
             goto error_no_tpm;
         tpm_running = true;
     }
@@ -500,6 +551,10 @@ int swtpm_main(int argc, char **argv, const char *prgname, const char *iface)
     if (create_seccomp_profile(false, seccomp_action) < 0)
         goto error_seccomp_profile;
 
+    if (daemonize) {
+        daemonize_finish();
+    }
+
     rc = mainLoop(&mlp, notify_fd[0]);
 
 error_seccomp_profile:
@@ -529,4 +584,9 @@ exit_failure:
     swtpm_cleanup(mlp.cc, server);
 
     exit(EXIT_FAILURE);
+
+exit_success:
+    swtpm_cleanup(mlp.cc, server);
+
+    exit(EXIT_SUCCESS);
 }
