@@ -136,7 +136,7 @@ static void usage(FILE *file, const char *prgname, const char *iface)
     "                 : use the given character device\n"
     "-f|--fd <fd>     : use the given character device file descriptor\n"
     "-d|--daemon      : daemonize the TPM\n"
-    "--ctrl type=[unixio|tcp][,path=<path>][,port=<port>[,bindaddr=address[,ifname=ifname]]][,fd=<filedescriptor|clientfd=<filedescriptor>][,mode=0...][,uid=uid][,gid=gid]\n"
+    "--ctrl type=[unixio|tcp][,path=<path>][,port=<port>[,bindaddr=address[,ifname=ifname]]][,fd=<filedescriptor|clientfd=<filedescriptor>][,mode=0...][,uid=uid][,gid=gid][,terminate]\n"
     "                 : TPM control channel using either UnixIO or TCP sockets;\n"
     "                   the path is only valid for Unixio channels; the port must\n"
     "                   be given in case the type is TCP; the TCP socket is bound\n"
@@ -150,6 +150,7 @@ static void usage(FILE *file, const char *prgname, const char *iface)
     "                   mode allows a user to set the file mode bits of a Unixio socket;\n"
     "                   the value must be given in octal number format\n"
     "                   uid and gid set the ownership of the Unixio socket's file;\n"
+    "                   terminate terminates on ctrl channel connection loss;\n"
     "--migration-key file=<path>|fd=<fd>[,mode=aes-cbc|aes-256-cbc][,format=hex|binary][,remove=[true|false]]\n"
     "                 : use an AES key for the encryption of the TPM's state\n"
     "                   when it is retrieved from the TPM via ioctls;\n"
@@ -182,6 +183,8 @@ static void usage(FILE *file, const char *prgname, const char *iface)
     "                   mode allows a user to set the file mode bits of the state files;\n"
     "                   the default mode is 0640;\n"
     "-r|--runas <user>: change to the given user\n"
+    "-R|--chroot <path>\n"
+    "                 : chroot to the given directory at startup\n"
 #ifdef WITH_VTPM_PROXY
     "--vtpm-proxy     : spawn a Linux vTPM proxy driver device and read TPM\n"
 #endif
@@ -189,12 +192,14 @@ static void usage(FILE *file, const char *prgname, const char *iface)
     "--locality [reject-locality-4][,allow-set-locality]\n"
     "                 : reject-locality-4: reject any command in locality 4\n"
     "                   allow-set-locality: accept SetLocality command\n"
-    "--flags [not-need-init][,startup-clear|startup-state|startup-deactivated|startup-none]\n"
+    "--flags [not-need-init][,startup-clear|startup-state|startup-deactivated|startup-none][,disable-auto-shutdown]\n"
     "                 : not-need-init: commands can be sent without needing to\n"
     "                   send an INIT via control channel; not needed when using\n"
     "                   --vtpm-proxy\n"
     "                   startup-...: send Startup command with this type;\n"
     "                   when --vtpm-proxy is used, startup-clear is used\n"
+    "                   disable-auto-shutdown disables automatic sending of\n"
+    "                   TPM2_Shutdown before TPM 2 reset or swtpm termination;\n"
     "--tpm2           : choose TPM2 functionality\n"
 #ifdef WITH_SECCOMP
 # ifndef SCMP_ACT_LOG
@@ -205,7 +210,11 @@ static void usage(FILE *file, const char *prgname, const char *iface)
     "                 : Choose the action of the seccomp profile when a\n"
     "                   blacklisted syscall is executed; default is kill\n"
 #endif
-    "--print-capabilites\n"
+    "--migration [incoming][,release-lock-outgoing]\n"
+    "                 : Incoming migration defers locking of storage backend\n"
+    "                   until the TPM state is received; release-lock-outgoing\n"
+    "                   releases the storage lock on outgoing migration\n"
+    "--print-capabilities\n"
     "                 : print capabilities and terminate\n"
     "--print-states\n"
     "                 : print existing TPM states and terminate\n"
@@ -274,6 +283,10 @@ int swtpm_chardev_main(int argc, char **argv, const char *prgname, const char *i
         .locality_flags = 0,
         .tpmversion = TPMLIB_TPM_VERSION_1_2,
         .startupType = _TPM_ST_NONE,
+        .lastCommand = TPM_ORDINAL_NONE,
+        .disable_auto_shutdown = false,
+        .incoming_migration = false,
+        .storage_locked = false,
     };
     unsigned long val;
     char *end_ptr;
@@ -286,7 +299,9 @@ int swtpm_chardev_main(int argc, char **argv, const char *prgname, const char *i
     char *ctrlchdata = NULL;
     char *flagsdata = NULL;
     char *seccompdata = NULL;
+    char *migrationdata = NULL;
     char *runas = NULL;
+    char *chroot = NULL;
 #ifdef WITH_VTPM_PROXY
     bool use_vtpm_proxy = false;
 #endif
@@ -303,6 +318,7 @@ int swtpm_chardev_main(int argc, char **argv, const char *prgname, const char *i
         {"chardev"   , required_argument, 0, 'c'},
         {"fd"        , required_argument, 0, 'f'},
         {"runas"     , required_argument, 0, 'r'},
+        {"chroot"    , required_argument, 0, 'R'},
         {"locality"  , required_argument, 0, 'L'},
         {"log"       , required_argument, 0, 'l'},
         {"key"       , required_argument, 0, 'k'},
@@ -318,6 +334,7 @@ int swtpm_chardev_main(int argc, char **argv, const char *prgname, const char *i
 #ifdef WITH_SECCOMP
         {"seccomp"   , required_argument, 0, 'S'},
 #endif
+        {"migration" , required_argument, 0, 'i'},
         {"print-capabilities"
                      ,       no_argument, 0, 'a'},
         {"print-states",     no_argument, 0, 'e'},
@@ -327,7 +344,7 @@ int swtpm_chardev_main(int argc, char **argv, const char *prgname, const char *i
     log_set_prefix("swtpm: ");
 
     while (TRUE) {
-        opt = getopt_long(argc, argv, "dhc:f:r:", longopts, &longindex);
+        opt = getopt_long(argc, argv, "dhc:f:r:R:", longopts, &longindex);
 
         if (opt == -1)
             break;
@@ -440,6 +457,10 @@ int swtpm_chardev_main(int argc, char **argv, const char *prgname, const char *i
             runas = optarg;
             break;
 
+        case 'R':
+            chroot = optarg;
+            break;
+
 #ifdef WITH_VTPM_PROXY
         case 'v':
             use_vtpm_proxy = true;
@@ -450,6 +471,10 @@ int swtpm_chardev_main(int argc, char **argv, const char *prgname, const char *i
             seccompdata = optarg;
             break;
 
+        case 'i':
+            migrationdata = optarg;
+            break;
+
         default:
             usage(stderr, prgname, iface);
             exit(EXIT_FAILURE);
@@ -462,6 +487,11 @@ int swtpm_chardev_main(int argc, char **argv, const char *prgname, const char *i
         exit(EXIT_FAILURE);
     }
 
+    if (chroot) {
+        if (do_chroot(chroot) < 0)
+            exit(EXIT_FAILURE);
+    }
+
     /* change process ownership before accessing files */
     if (runas) {
         if (change_process_owner(runas) < 0)
@@ -486,16 +516,12 @@ int swtpm_chardev_main(int argc, char **argv, const char *prgname, const char *i
          * Choose the TPM version so that getting/setting buffer size works.
          * Ignore failure, for backward compatibility when TPM 1.2 is disabled.
          */
-        TPMLIB_ChooseTPMVersion(mlp.tpmversion);
-        ret = capabilities_print_json(false);
+        ret = capabilities_print_json(false, mlp.tpmversion);
         exit(ret ? EXIT_FAILURE : EXIT_SUCCESS);
     }
 
-    if (TPMLIB_ChooseTPMVersion(mlp.tpmversion) != TPM_SUCCESS) {
-        logprintf(STDERR_FILENO,
-                  "Error: Could not choose TPM version.\n");
+    if (tpmlib_choose_tpm_version(mlp.tpmversion) != TPM_SUCCESS)
         exit(EXIT_FAILURE);
-    }
 
     tpmstate_set_version(mlp.tpmversion);
 
@@ -525,7 +551,7 @@ int swtpm_chardev_main(int argc, char **argv, const char *prgname, const char *i
         exit(EXIT_FAILURE);
     }
 
-    if (handle_ctrlchannel_options(ctrlchdata, &mlp.cc) < 0) {
+    if (handle_ctrlchannel_options(ctrlchdata, &mlp.cc, &mlp.flags) < 0) {
         goto exit_failure;
     }
 
@@ -535,7 +561,9 @@ int swtpm_chardev_main(int argc, char **argv, const char *prgname, const char *i
         handle_tpmstate_options(tpmstatedata) < 0 ||
         handle_seccomp_options(seccompdata, &seccomp_action) < 0 ||
         handle_flags_options(flagsdata, &need_init_cmd,
-                             &mlp.startupType) < 0) {
+                             &mlp.startupType, &mlp.disable_auto_shutdown) < 0 ||
+        handle_migration_options(migrationdata, &mlp.incoming_migration,
+                                 &mlp.release_lock_outgoing) < 0) {
         goto exit_failure;
     }
 
@@ -558,7 +586,9 @@ int swtpm_chardev_main(int argc, char **argv, const char *prgname, const char *i
         goto error_no_tpm;
 
     if (!need_init_cmd) {
-        if ((rc = tpmlib_start(0, mlp.tpmversion)))
+        mlp.storage_locked = !mlp.incoming_migration;
+
+        if ((rc = tpmlib_start(0, mlp.tpmversion, mlp.storage_locked)))
             goto error_no_tpm;
         tpm_running = true;
     }
