@@ -207,6 +207,31 @@ void SWTPM_NVRAM_Shutdown(void)
     memset(&migrationkey, 0, sizeof(migrationkey));
 }
 
+TPM_RESULT SWTPM_NVRAM_Lock_Storage(unsigned int retries)
+{
+    const char *backend_uri;
+
+    if (!g_nvram_backend_ops)
+        return TPM_RETRY;
+
+    backend_uri = tpmstate_get_backend_uri();
+    if (!backend_uri) {
+        logprintf(STDERR_FILENO,
+                  "SWTPM_NVRAM_Lock: Missing backend URI.\n");
+        return TPM_FAIL;
+    }
+    if (g_nvram_backend_ops->lock)
+        return g_nvram_backend_ops->lock(backend_uri, retries);
+
+    return TPM_SUCCESS;
+}
+
+void SWTPM_NVRAM_Unlock(void)
+{
+    if (g_nvram_backend_ops->unlock)
+        g_nvram_backend_ops->unlock();
+}
+
 /* SWTPM_NVRAM_GetFilenameForName() constructs a file name from the name.
  * A temporary filename used to write to may be created. It should be rename()'d to
  * the non-temporary filename.
@@ -265,7 +290,6 @@ SWTPM_NVRAM_LoadData(unsigned char **data,     /* freed by caller */
                      const char *name)
 {
     TPM_RESULT    rc = 0;
-    int           irc;
     unsigned char *decrypt_data = NULL;
     uint32_t      decrypt_length;
     uint32_t      dataoffset = 0;
@@ -283,18 +307,17 @@ SWTPM_NVRAM_LoadData(unsigned char **data,     /* freed by caller */
                                        backend_uri);
     }
 
+    /* this function needs to return the plain data -- no tlv headers */
     if (rc == 0) {
-        /* this function needs to return the plain data -- no tlv headers */
-
-        /* try to get a header from it -- old files may not have one */
-        irc = SWTPM_NVRAM_CheckHeader(*data, *length, &dataoffset,
+        rc = SWTPM_NVRAM_CheckHeader(*data, *length, &dataoffset,
                                       &hdrflags, &hdrversion, true);
-        /* valid header -- this one can only be version 2 or later */
-        if (irc) {
-            hdrversion = 1; /* no header -- payload was written like vers. 1 */
-            hdrflags = 0;
-        }
+        if (rc != 0)
+            logprintf(STDERR_FILENO,
+                      "SWTPM_NVRAM_LoadData: Error from SWTPM_NVRAM_CheckHeader "
+                      "rc = %d\n", rc);
+    }
 
+    if (rc == 0) {
         rc = SWTPM_NVRAM_GetDecryptedData(&filekey,
                                           &decrypt_data, &decrypt_length,
                                           *data + dataoffset,
@@ -310,19 +333,17 @@ SWTPM_NVRAM_LoadData(unsigned char **data,     /* freed by caller */
             logprintf(STDERR_FILENO,
                       "SWTPM_NVRAM_LoadData: Error from SWTPM_NVRAM_GetDecryptedData "
                       "rc = %d\n", rc);
-
-        if (rc == 0) {
-            TPM_DEBUG(" SWTPM_NVRAM_LoadData: Decrypted %u bytes of "
-                      "data to %u bytes.\n",
-                      *length, decrypt_length);
-            free(*data);
-            *data = decrypt_data;
-            *length = decrypt_length;
-        }
     }
 
-    if (rc != 0) {
-        free(*data);
+    free(*data);
+
+    if (rc == 0) {
+        TPM_DEBUG(" SWTPM_NVRAM_LoadData: Decrypted %u bytes of "
+                  "data to %u bytes.\n",
+                  *length, decrypt_length);
+        *data = decrypt_data;
+        *length = decrypt_length;
+    } else {
         *data = NULL;
     }
 
@@ -865,7 +886,7 @@ SWTPM_NVRAM_DecryptData(const encryptionkey *key,
                         uint16_t tag_ivec, uint16_t hdrflags,
                         uint16_t flag_256bitkey)
 {
-    TPM_RESULT rc = 0;
+    TPM_RESULT rc = TPM_FAIL;
     unsigned char *tmp_data = NULL;
     uint32_t tmp_length = 0;
     tlv_data td[2];
@@ -873,63 +894,62 @@ SWTPM_NVRAM_DecryptData(const encryptionkey *key,
     uint32_t ivec_length = 0;
     size_t keylen;
 
-    if (key->symkey.userKeyLength > 0) {
-        switch (key->data_encmode) {
-        case ENCRYPTION_MODE_UNKNOWN:
-            rc = TPM_BAD_MODE;
-            break;
-        case ENCRYPTION_MODE_AES_CBC:
-            switch (hdrversion) {
-            case 1:
-                rc = SWTPM_SymmetricKeyData_Decrypt(&tmp_data,
-                                                    &tmp_length,
-                                                    data, length,
-                                                    &key->symkey,
-                                                    NULL, 0);
-                if (rc == 0) {
-                    rc = SWTPM_CheckHash(tmp_data, tmp_length,
-                                         decrypt_data, decrypt_length);
-                }
-            break;
-            case 2:
-                keylen = (hdrflags & flag_256bitkey)
-                          ? SWTPM_AES256_BLOCK_SIZE : SWTPM_AES128_BLOCK_SIZE;
-                if (keylen != key->symkey.userKeyLength) {
-                    logprintf(STDERR_FILENO,
-                              "Wrong decryption key. Need %zu bit key.\n",
-                              keylen * 8);
-                    rc = TPM_BAD_KEY_PROPERTY;
-                    break;
-                }
+    if (key->symkey.userKeyLength == 0)
+        return rc;
 
-                if (!tlv_data_find_tag(data, length, TAG_HMAC, &td[0]) ||
-                    !tlv_data_find_tag(data, length, tag_encrypted_data,
-                                       &td[1])) {
-                    logprintf(STDERR_FILENO,
-                              "Could not find HMAC or encrypted data (tag %u) "
-                              "in byte stream.\n", tag_encrypted_data);
-                    rc = TPM_FAIL;
-                    break;
-                }
-                /* get the IV, if there is one */
-                SWTPM_GetIvec(data, length, &ivec, &ivec_length, tag_ivec);
-
-                rc = SWTPM_CheckHMAC(&td[0], &td[1], &key->symkey,
-                                     ivec, ivec_length);
-                if (rc == 0) {
-                    rc = SWTPM_SymmetricKeyData_Decrypt(decrypt_data,
-                                                        decrypt_length,
-                                                        td[1].u.const_ptr,
-                                                        td[1].tlv.length,
-                                                        &key->symkey,
-                                                        ivec, ivec_length);
-                }
-            break;
-            default:
-                rc = TPM_FAIL;
+    switch (key->data_encmode) {
+    case ENCRYPTION_MODE_UNKNOWN:
+        rc = TPM_BAD_MODE;
+        break;
+    case ENCRYPTION_MODE_AES_CBC:
+        switch (hdrversion) {
+        case 1:
+            rc = SWTPM_SymmetricKeyData_Decrypt(&tmp_data,
+                                                &tmp_length,
+                                                data, length,
+                                                &key->symkey,
+                                                NULL, 0);
+            if (rc == 0) {
+                rc = SWTPM_CheckHash(tmp_data, tmp_length,
+                                     decrypt_data, decrypt_length);
             }
-            free(tmp_data);
+        break;
+        case 2:
+            keylen = (hdrflags & flag_256bitkey)
+                      ? SWTPM_AES256_BLOCK_SIZE : SWTPM_AES128_BLOCK_SIZE;
+            if (keylen != key->symkey.userKeyLength) {
+                logprintf(STDERR_FILENO,
+                          "Wrong decryption key. Need %zu bit key.\n",
+                          keylen * 8);
+                rc = TPM_BAD_KEY_PROPERTY;
+                break;
+            }
+
+            if (!tlv_data_find_tag(data, length, TAG_HMAC, &td[0]) ||
+                !tlv_data_find_tag(data, length, tag_encrypted_data,
+                                   &td[1])) {
+                logprintf(STDERR_FILENO,
+                          "Could not find HMAC or encrypted data (tag %u) "
+                          "in byte stream.\n", tag_encrypted_data);
+                rc = TPM_FAIL;
+                break;
+            }
+            /* get the IV, if there is one */
+            SWTPM_GetIvec(data, length, &ivec, &ivec_length, tag_ivec);
+
+            rc = SWTPM_CheckHMAC(&td[0], &td[1], &key->symkey,
+                                 ivec, ivec_length);
+            if (rc == 0) {
+                rc = SWTPM_SymmetricKeyData_Decrypt(decrypt_data,
+                                                    decrypt_length,
+                                                    td[1].u.const_ptr,
+                                                    td[1].tlv.length,
+                                                    &key->symkey,
+                                                    ivec, ivec_length);
+            }
+        break;
         }
+        free(tmp_data);
     }
 
     return rc;
