@@ -660,7 +660,8 @@ static int do_load_state_blob(int fd, bool is_chardev, const char *blobtype,
     return 0;
 }
 
-static int change_fd_flags(int fd, int flags_to_clear, int flags_to_set) {
+static int change_fd_flags(int fd, int flags_to_clear, int flags_to_set,
+                           int *o_flags) {
     int n;
     int orig_flags = fcntl(fd, F_GETFL, 0);
 
@@ -675,7 +676,15 @@ static int change_fd_flags(int fd, int flags_to_clear, int flags_to_set) {
             return -1;
         }
     }
-    return orig_flags;
+    if (o_flags)
+        *o_flags = orig_flags;
+
+    return 0;
+}
+
+static int set_fd_flags(int fd, int flags_to_set)
+{
+    return change_fd_flags(fd, ~0, flags_to_set, NULL);
 }
 
 /* Create a connection by setting the given file descriptor to non-blocking.
@@ -688,11 +697,14 @@ static int connect_nonblock(int fd, const struct sockaddr *addr,
 {
     int n, sockerr;
     socklen_t optlen = sizeof(sockerr);
+    int orig_flags = 0;
+
 #if !defined(__CYGWIN__)
-    int orig_flags = change_fd_flags(fd, 0, O_NONBLOCK);
+    if (change_fd_flags(fd, 0, O_NONBLOCK, &orig_flags) < 0)
 #else
-    int orig_flags = change_fd_flags(fd, 0, 0);
+    if (change_fd_flags(fd, 0, 0, &orig_flags) < 0)
 #endif
+        return -1;
 
     n = connect(fd, addr, addrlen);
     /* n < 0: it must fail with EAGAIN (Unix socket) and then we have to poll
@@ -726,7 +738,7 @@ static int connect_nonblock(int fd, const struct sockaddr *addr,
             return -1;
         }
     }
-    return change_fd_flags(fd, ~0, orig_flags);
+    return set_fd_flags(fd, orig_flags);
 }
 
 static int open_connection(const char *devname, char *tcp_hostname,
@@ -798,7 +810,8 @@ static int open_connection(const char *devname, char *tcp_hostname,
     return fd;
 }
 
-static int parse_tcp_optarg(char *optarg, char **tcp_hostname, int *tcp_port)
+static int parse_tcp_optarg(char *optarg, char **tcp_hostname,
+                            unsigned int *tcp_port)
 {
     char *pos = strrchr(optarg, ':');
     int n;
@@ -898,6 +911,8 @@ static void usage(const char *prgname)
 "                        size; get minimum and maximum supported sizes\n"
 "--info <flags>        : get TPM implementation specific information;\n"
 "                        flags must be an integer value\n"
+"--lock-storage <n>    : lock the storage after it was unlocked; retry\n"
+"                        n times with 10ms delay in between\n"
 "--version             : display version and exit\n"
 "--help                : display help screen and exit\n"
 "\n"
@@ -906,7 +921,7 @@ static void usage(const char *prgname)
 
 int main(int argc, char *argv[])
 {
-    int fd, n;
+    int fd = -1, n;
     ptm_est est;
     ptm_reset_est reset_est;
     ptm_loc loc;
@@ -916,6 +931,7 @@ int main(int argc, char *argv[])
     ptm_getconfig cfg;
     ptm_setbuffersize psbs;
     ptm_getinfo pgi;
+    ptm_lockstorage pls;
     char *tmp;
     size_t buffersize = 0;
     static struct option long_options[] = {
@@ -938,6 +954,7 @@ int main(int argc, char *argv[])
         {"load", required_argument, NULL, 'L'},
         {"version", no_argument, NULL, 'V'},
         {"info", required_argument, NULL, 'I'},
+        {"lock-storage", required_argument, NULL, 'o'},
         {"help", no_argument, NULL, 'H'},
         {NULL, 0, NULL, 0},
     };
@@ -948,7 +965,8 @@ int main(int argc, char *argv[])
     char *tcp_hostname = NULL;
     unsigned int locality = 0;
     unsigned int tpmbuffersize = 0;
-    int tcp_port = -1;
+    unsigned int tcp_port = 0;
+    unsigned int retries;
     bool is_chardev;
     unsigned long int info_flags = 0;
     char *endptr = NULL;
@@ -966,6 +984,7 @@ int main(int argc, char *argv[])
             tpm_device = optarg;
             break;
         case 'T':
+            free(tcp_hostname);
             if (parse_tcp_optarg(optarg, &tcp_hostname, &tcp_port) < 0)
                 return EXIT_FAILURE;
             break;
@@ -1040,6 +1059,14 @@ int main(int argc, char *argv[])
                 goto exit;
             }
             break;
+        case 'o':
+            command = argv[optind - 2];
+            if (sscanf(argv[optind - 1], "%u", &retries) != 1) {
+                fprintf(stderr, "Could not parse lock retries from %s.\n",
+                        argv[optind - 1]);
+                goto exit;
+            }
+            break;
         case 'V':
             versioninfo();
             ret = EXIT_SUCCESS;
@@ -1287,24 +1314,63 @@ int main(int argc, char *argv[])
                devtoh32(is_chardev, psbs.u.resp.minsize),
                devtoh32(is_chardev, psbs.u.resp.maxsize));
     } else if (!strcmp(command, "--info")) {
+        char buffer[sizeof(pgi.u.resp.buffer) + 1];
+        uint32_t bytes_read = 0;
+        uint32_t len;
+
         memset(&pgi, 0, sizeof(pgi));
-        pgi.u.req.flags = htodev64(is_chardev, info_flags);
-        pgi.u.req.offset = htodev64(is_chardev, 0);
-        n = ctrlcmd(fd, PTM_GET_INFO, &pgi,
-                    sizeof(pgi.u.req), sizeof(pgi.u.resp));
+        do {
+            pgi.u.req.flags = htodev64(is_chardev, info_flags);
+            pgi.u.req.offset = htodev32(is_chardev, bytes_read);
+            n = ctrlcmd(fd, PTM_GET_INFO, &pgi,
+                        sizeof(pgi.u.req), sizeof(pgi.u.resp));
+            if (n < 0) {
+                fprintf(stderr,
+                        "Could not execute PTM_GET_INFO: %s\n",
+                        strerror(errno));
+                goto exit;
+            }
+            res = devtoh32(is_chardev, pgi.u.resp.tpm_result);
+            if (res != 0) {
+                fprintf(stderr,
+                        "TPM result from PTM_GET_INFO: 0x%x\n", res);
+                goto exit;
+            }
+
+            len = devtoh32(is_chardev, pgi.u.resp.length);
+            if (len > sizeof(pgi.u.resp.buffer)) {
+                fprintf(stderr,
+                        "TPM returned an invalid length of %u, expected maximum %zu\n",
+                        len, sizeof(pgi.u.resp.buffer));
+                goto exit;
+            }
+            bytes_read += len;
+
+            memcpy(buffer, pgi.u.resp.buffer, len);
+            buffer[len] = 0;
+            printf("%s", buffer);
+        } while (bytes_read < devtoh32(is_chardev, pgi.u.resp.totlength));
+        printf("\n");
+    } else if (!strcmp(command, "--lock-storage")) {
+        memset(&pls, 0, sizeof(pls));
+        pls.u.req.retries = htodev32(is_chardev, 0);
+
+        n = ctrlcmd(fd, PTM_LOCK_STORAGE, &pls,
+                    sizeof(pls.u.req), sizeof(pls.u.resp));
         if (n < 0) {
             fprintf(stderr,
-                    "Could not execute PTM_GET_INFO: %s\n",
-                    strerror(errno));
+                    "Could not execute PTM_LOCK_STORAGE: "
+                    "%s\n", strerror(errno));
             goto exit;
         }
-        res = devtoh32(is_chardev, pgi.u.resp.tpm_result);
-        if (res != 0) {
+        res = devtoh32(is_chardev, pls.u.resp.tpm_result);
+        if (devtoh32(is_chardev, res) != 0) {
             fprintf(stderr,
-                    "TPM result from PTM_GET_INFO: 0x%x\n", res);
+                    "TPM result from PTM_LOCK_STORAGE: 0x%x\n",
+                    devtoh32(is_chardev, res));
             goto exit;
         }
-        printf("%s\n", pgi.u.resp.buffer);
+
     } else {
         usage(argv[0]);
         goto exit;
@@ -1314,5 +1380,8 @@ int main(int argc, char *argv[])
 
 exit:
     free(tcp_hostname);
+    if (fd >= 0)
+        close(fd);
+
     return ret;
 }
