@@ -1,6 +1,6 @@
 /* SPDX-License-Identifier: BSD-3-Clause */
 /*
- * swtpm_localca.c: A tool for creating TPM 1.2 and TPM 2 certificates localy or using pkcs11
+ * swtpm_localca.c: A tool for creating TPM 1.2 and TPM 2 certificates locally or using pkcs11
  *
  * Author: Stefan Berger, stefanb@linux.ibm.com
  *
@@ -24,8 +24,10 @@
 
 #include <glib.h>
 
+#include <gmp.h>
+
+#include "swtpm_conf.h"
 #include "swtpm_utils.h"
-#include "swtpm_localca_conf.h"
 #include "swtpm_localca_utils.h"
 
 #define SETUP_TPM2_F    1
@@ -253,7 +255,7 @@ error:
  * This function returns 1  on error, 2 if the ECC parameters could be extracted
  * and 0 if no parameters could be extracted (likely a modulus).
  */
-static gboolean extract_ecc_params(const gchar *ekparams, gchar **ecc_x, gchar **ecc_y, gchar **ecc_curveid)
+static gboolean extract_ecc_params(const gchar *key_params, gchar **ecc_x, gchar **ecc_y, gchar **ecc_curveid)
 {
     regmatch_t pmatch[5];
     regex_t preg;
@@ -266,13 +268,13 @@ static gboolean extract_ecc_params(const gchar *ekparams, gchar **ecc_x, gchar *
     }
 
     ret = 0;
-    if (regexec(&preg, ekparams, 5, pmatch, 0) == 0) {
-        *ecc_x = g_strndup(&ekparams[pmatch[1].rm_so],
+    if (regexec(&preg, key_params, 5, pmatch, 0) == 0) {
+        *ecc_x = g_strndup(&key_params[pmatch[1].rm_so],
                            pmatch[1].rm_eo - pmatch[1].rm_so);
-        *ecc_y = g_strndup(&ekparams[pmatch[2].rm_so],
+        *ecc_y = g_strndup(&key_params[pmatch[2].rm_so],
                            pmatch[2].rm_eo - pmatch[2].rm_so);
         if (pmatch[4].rm_so > 0 && pmatch[4].rm_eo > 0)
-            *ecc_curveid = g_strndup(&ekparams[pmatch[4].rm_so],
+            *ecc_curveid = g_strndup(&key_params[pmatch[4].rm_so],
                                      pmatch[4].rm_eo - pmatch[4].rm_so);
         ret = 2;
     }
@@ -282,16 +284,36 @@ static gboolean extract_ecc_params(const gchar *ekparams, gchar **ecc_x, gchar *
     return ret;
 }
 
+/* Create a random ASCII decimal number of given length.
+ * The buffer is not NUL terminated.
+ */
+static void get_random_serial(char *buffer, size_t length)
+{
+    GRand *grand = g_rand_new();
+    size_t i;
+
+    buffer[0] = '0' + g_rand_int_range(grand, 1, 9);
+    for (i = 1; i < length; i++)
+        buffer[i] = '0' + g_rand_int_range(grand, 0, 9);
+
+    g_rand_free(grand);
+}
+
 /* Get the next serial number from the certserial file; if it contains
- * a non-numeric content start over with serial number '1'.
+ * a non-numeric content start over with a random 20 digit serial number.
+ * Up to 20 bytes of serial number are supported. The max.
+ * serial number is decimal: 1461501637330902918203684832716283019655932542975
+ * This decimal number is 49 digits long.
+ * This function will write back the used serial number that the next
+ * caller must increase by '1' to be allowed to use it.
  */
 static int get_next_serial(const gchar *certserial, const gchar *lockfile,
                            gchar **serial_str)
 {
     g_autofree gchar *buffer = NULL;
+    char serialbuffer[50];
     size_t buffer_len;
-    unsigned long long serial, serial_n;
-    char *endptr = NULL;
+    mpz_t serial;
     int lockfd;
     int ret = 1;
 
@@ -299,25 +321,43 @@ static int get_next_serial(const gchar *certserial, const gchar *lockfile,
     if (lockfd < 0)
         return 1;
 
-    if (access(certserial, R_OK) != 0)
-        write_file(certserial, (unsigned char *)"1", 1);
+    if (access(certserial, R_OK) != 0) {
+        get_random_serial(serialbuffer, 20);
+        write_file(certserial, (unsigned char *)serialbuffer, 20);
+    }
     if (read_file(certserial, &buffer, &buffer_len) != 0)
         goto error;
 
-    if (buffer_len > 0) {
-        serial = strtoull(buffer, &endptr, 10);
-        if (*endptr == '\0') {
-            serial_n = serial + 1;
-        } else {
-            serial_n = 1;
-        }
+    mpz_init(serial);
+
+    if (buffer_len > 0 && buffer_len <= 49) {
+        memcpy(serialbuffer, buffer, buffer_len);
+        serialbuffer[buffer_len] = 0;
+
+        if (gmp_sscanf(serialbuffer, "%Zu", serial) != 1)
+            goto new_serial;
+        mpz_add_ui(serial, serial, 1);
+
+        if ((mpz_sizeinbase(serial, 2) + 7) / 8 > 20)
+            goto new_serial;
+
+        if (gmp_snprintf(serialbuffer,
+                         sizeof(serialbuffer),
+                         "%Zu", serial) >= (int)sizeof(serialbuffer))
+            goto new_serial;
     } else {
-        serial_n = 1;
+new_serial:
+        /* start with random serial number */
+        buffer_len = 20;
+        get_random_serial(serialbuffer, buffer_len);
+        serialbuffer[buffer_len] = 0;
     }
-    *serial_str = g_strdup_printf("%llu", serial_n);
+    *serial_str = g_strdup(serialbuffer);
     write_file(certserial, (unsigned char *)*serial_str, strlen(*serial_str));
     ret = 0;
 
+    mpz_clear(serial);
+
 error:
     unlock_file(lockfd);
 
@@ -326,7 +366,7 @@ error:
 
 /* Create a TPM 1.2 or TPM 2 EK or platform cert */
 static int create_cert(unsigned long flags, const gchar *typ, const gchar *directory,
-                       gchar *ekparams, const gchar *vmid, gchar **tpm_spec_params,
+                       gchar *key_params, const gchar *vmid, gchar **tpm_spec_params,
                        gchar **tpm_attr_params, const gchar *signkey,
                        const gchar *signkey_password, const gchar *issuercert,
                        const gchar *parentkey_password, gchar **swtpm_cert_env,
@@ -370,7 +410,7 @@ static int create_cert(unsigned long flags, const gchar *typ, const gchar *direc
     if (get_next_serial(certserial, lockfile, &serial_str) != 0)
         return 1;
 
-    /* try to read the optsfile */
+    /* try to read the optsfile - failure to read is fine */
     read_file_lines(optsfile, &optsfile_lines);
 
     /* split each line from the optsfile and add the stripped parameters to options */
@@ -409,7 +449,7 @@ static int create_cert(unsigned long flags, const gchar *typ, const gchar *direc
             options = concat_arrays(options, (gchar *[]){"--decryption", NULL}, TRUE);
     }
 
-    switch (extract_ecc_params(ekparams, &ecc_x, &ecc_y, &ecc_curveid)) {
+    switch (extract_ecc_params(key_params, &ecc_x, &ecc_y, &ecc_curveid)) {
     case 1:
         goto error;
     case 2:
@@ -428,7 +468,7 @@ static int create_cert(unsigned long flags, const gchar *typ, const gchar *direc
         break;
     case 0:
         keyparams = concat_arrays((gchar *[]){
-                                      "--modulus", ekparams,
+                                      "--modulus", key_params,
                                       NULL},
                                    NULL, FALSE);
         break;
@@ -540,6 +580,7 @@ static void usage(const char *prgname)
         "\n"
         "--type type           The type of certificate to create: 'ek' or 'platform'\n"
         "--ek key-param        The modulus of an RSA key or x=...,y=,... for an EC key\n"
+        "--key key-param       Alias for --ek\n"
         "--dir directory       The directory to write the resulting certificate into\n"
         "--vmid vmid           The ID of the virtual machine\n"
         "--optsfile file       A file containing options to pass to swtpm_cert\n"
@@ -570,6 +611,7 @@ int main(int argc, char *argv[])
     static const struct option long_options[] = {
         {"type", required_argument, NULL, 't'},
         {"ek", required_argument, NULL, 'e'},
+        {"key", required_argument, NULL, 'e'}, /* alias for --ek */
         {"dir", required_argument, NULL, 'd'},
         {"vmid", required_argument, NULL, 'v'},
         {"optsfile", required_argument, NULL, 'o'},
@@ -585,6 +627,7 @@ int main(int argc, char *argv[])
         {"allow-signing", no_argument, NULL, 'i'},
         {"decryption", no_argument, NULL, 'y'},
         {"help", no_argument, NULL, 'h'},
+        {NULL, 0, NULL, 0},
     };
     g_autofree gchar *default_options_file = NULL;
     g_autofree gchar *default_config_file = NULL;
@@ -592,7 +635,7 @@ int main(int argc, char *argv[])
     g_autofree gchar *configfile = NULL;
     unsigned long flags = 0;
     g_autofree gchar *typ =g_strdup("");
-    g_autofree gchar *ekparams = g_strdup("");
+    g_autofree gchar *key_params = g_strdup("");
     g_autofree gchar *directory = g_strdup("."); /* default to current directory */
     g_autofree gchar *vmid = NULL;
     g_autofree gchar *lockfile = NULL;
@@ -610,6 +653,8 @@ int main(int argc, char *argv[])
     struct stat statbuf;
     int ret = 1;
 
+    setvbuf(stdout, 0, _IONBF, 0);
+
     if (init(&default_options_file, &default_config_file) < 0)
         goto error;
     optsfile = g_strdup(default_options_file);
@@ -622,9 +667,9 @@ int main(int argc, char *argv[])
             g_free(typ);
             typ = g_strdup(optarg);
             break;
-        case 'e': /* --ek */
-            g_free(ekparams);
-            ekparams = g_strdup(optarg);
+        case 'e': /* --ek or --key */
+            g_free(key_params);
+            key_params = g_strdup(optarg);
             break;
         case 'd': /* --dir */
             g_free(directory);
@@ -861,7 +906,7 @@ int main(int argc, char *argv[])
             goto error;
     }
 
-    ret = create_cert(flags, typ, directory, ekparams, vmid, tpm_spec_params, tpm_attr_params,
+    ret = create_cert(flags, typ, directory, key_params, vmid, tpm_spec_params, tpm_attr_params,
                       signkey, signkey_password, issuercert, parentkey_password, swtpm_cert_env,
                       certserial, lockfile, optsfile);
 
