@@ -10,6 +10,8 @@
 #include "config.h"
 
 #include <errno.h>
+#include <poll.h>
+#include <stdbool.h>
 #include <stdio.h>
 #include <stdint.h>
 #include <string.h>
@@ -18,6 +20,7 @@
 #include <sys/stat.h>
 #include <sys/wait.h>
 #include <unistd.h>
+#include <fcntl.h>
 
 #include <glib.h>
 
@@ -42,6 +45,8 @@
 #define AS4BE(VAL) AS2BE((VAL) >> 16), AS2BE(VAL)
 #define AS8BE(VAL) AS4BE((VAL) >> 32), AS4BE(VAL)
 
+#define CMD_DURATION_SHORT  2000 /* ms */
+
 struct tpm_req_header {
     uint16_t tag;
     uint32_t size;
@@ -54,6 +59,20 @@ struct tpm_resp_header {
     uint32_t errcode;
 } __attribute__((packed));
 
+/* Close the ctrl and data file descriptors that were passed to the swtpm process.
+ * If 'all' is true then also close the ones not passed to the process.
+ */
+static void swtpm_close_comm(struct swtpm *self, bool all)
+{
+    if (all)
+        SWTPM_CLOSE(self->data_fds[0]);
+    SWTPM_CLOSE(self->data_fds[1]);
+
+    if (all)
+        SWTPM_CLOSE(self->ctrl_fds[0]);
+    SWTPM_CLOSE(self->ctrl_fds[1]);
+}
+
 static int swtpm_start(struct swtpm *self)
 {
     g_autofree gchar *tpmstate = g_strdup_printf("backend-uri=%s", self->state_path);
@@ -66,12 +85,13 @@ static int swtpm_start(struct swtpm *self)
     struct stat statbuf;
     gboolean success;
     GError *error = NULL;
+    GSpawnFlags flags;
     unsigned ctr;
     int pidfile_fd;
     int ret = 1;
     char pidfile[] = "/tmp/.swtpm_setup.pidfile.XXXXXX";
 
-    pidfile_fd = mkstemp(pidfile);
+    pidfile_fd = g_mkstemp_full(pidfile, O_EXCL|O_CREAT, 0600);
     if (pidfile_fd < 0) {
         logerr(self->logfile, "Could not create pidfile: %s\n", strerror(errno));
         goto error_no_pidfile;
@@ -128,8 +148,16 @@ static int swtpm_start(struct swtpm *self)
     }
 #endif
 
-    success = g_spawn_async(NULL, argv, NULL,
-                            G_SPAWN_LEAVE_DESCRIPTORS_OPEN | G_SPAWN_STDOUT_TO_DEV_NULL | G_SPAWN_STDERR_TO_DEV_NULL,
+    flags = G_SPAWN_LEAVE_DESCRIPTORS_OPEN;
+    if (gl_LOGFILE) {
+        flags |= G_SPAWN_STDOUT_TO_DEV_NULL | G_SPAWN_STDERR_TO_DEV_NULL;
+    } else {
+#if GLIB_CHECK_VERSION(2, 74, 0)
+        flags |= G_SPAWN_CHILD_INHERITS_STDOUT | G_SPAWN_CHILD_INHERITS_STDERR;
+#endif
+    }
+
+    success = g_spawn_async(NULL, argv, NULL, flags,
                             NULL, NULL, &self->pid, &error);
     if (!success) {
         logerr(self->logfile, "Could not start swtpm: %s\n", error->message);
@@ -155,6 +183,8 @@ static int swtpm_start(struct swtpm *self)
     }
 
 error:
+    swtpm_close_comm(self, ret != 0);
+
     close(pidfile_fd);
     unlink(pidfile);
 
@@ -184,17 +214,7 @@ static void swtpm_stop(struct swtpm *self)
         self->pid = 0;
     }
 
-    if (self->ctrl_fds[0] >= 0) {
-        close(self->ctrl_fds[0]);
-        close(self->ctrl_fds[1]);
-        self->ctrl_fds[0] = self->ctrl_fds[1] = -1;
-    }
-
-    if (self->data_fds[0] >= 0) {
-        close(self->data_fds[0]);
-        close(self->data_fds[1]);
-        self->data_fds[0] = self->data_fds[1] = -1;
-    }
+    swtpm_close_comm(self, true);
 }
 
 /* Destroy a running swtpm instance */
@@ -206,7 +226,7 @@ static void swtpm_destroy(struct swtpm *self)
 /* Send a command to swtpm and receive the response either via control or data channel */
 static int transfer(struct swtpm *self, void *buffer, size_t buffer_len,
                     const char *cmdname, gboolean use_ctrl,
-                    void *respbuffer, size_t *respbuffer_len)
+                    void *respbuffer, size_t *respbuffer_len, int timeout_ms)
 {
     size_t offset;
     int sockfd;
@@ -214,6 +234,9 @@ static int transfer(struct swtpm *self, void *buffer, size_t buffer_len,
     unsigned char resp[4096];
     ssize_t resplen;
     uint32_t returncode;
+    struct pollfd fds = {
+        .events = POLLIN | POLLERR | POLLHUP,
+    };
 
     if (use_ctrl) {
         sockfd = self->ctrl_fds[0];
@@ -235,6 +258,14 @@ static int transfer(struct swtpm *self, void *buffer, size_t buffer_len,
         return 1;
     }
 
+    fds.fd = sockfd;
+    n = poll(&fds, 1, timeout_ms);
+    if (n != 1 || (fds.revents & POLLIN) == 0) {
+        logerr(self->logfile, "Could not receive response to %s from swtpm: %s\n",
+               cmdname, strerror(errno));
+        return 1;
+    }
+
     resplen = read(sockfd, resp, sizeof(resp));
     if (resplen < 0) {
         logerr(self->logfile, "Could not receive response to %s from swtpm: %s\n",
@@ -275,7 +306,8 @@ static int swtpm_ctrl_shutdown(struct swtpm *self)
 {
     uint32_t cmd = htobe32(CMD_SHUTDOWN);
 
-    return transfer(self, &cmd, sizeof(cmd), "CMD_SHUTDOWN", TRUE, NULL, 0);
+    return transfer(self, &cmd, sizeof(cmd), "CMD_SHUTDOWN", TRUE,
+                    NULL, NULL, CMD_DURATION_SHORT);
 }
 
 /* Get the TPM specification parameters over the control channel */
@@ -289,7 +321,8 @@ static int swtpm_ctrl_get_tpm_specs_and_attrs(struct swtpm *self, gchar **result
     int ret;
     uint32_t length;
 
-    ret = transfer(self, req, sizeof(req), "CMD_GET_INFO", TRUE, tpmresp, &tpmresp_len);
+    ret = transfer(self, req, sizeof(req), "CMD_GET_INFO", TRUE,
+                   tpmresp, &tpmresp_len, CMD_DURATION_SHORT);
     if (ret != 0)
         return 1;
 
@@ -332,6 +365,7 @@ static const struct swtpm_cops swtpm_cops = {
 #define TPM2_CC_NV_WRITE       0x00000137
 #define TPM2_CC_NV_WRITELOCK   0x00000138
 #define TPM2_CC_SHUTDOWN       0x00000145
+#define TPM2_CC_FLUSHCONTEXT   0x00000165
 #define TPM2_CC_GETCAPABILITY  0x0000017a
 
 #define TPM2_SU_CLEAR        0x0000
@@ -386,6 +420,10 @@ static const struct swtpm_cops swtpm_cops = {
 #define TPM2_EK_ECC_SECP384R1_HANDLE 0x81010016
 #define TPM2_SPK_HANDLE              0x81000001
 
+#define TPM2_DURATION_SHORT     2000 /* ms */
+#define TPM2_DURATION_MEDIUM    7500 /* ms */
+#define TPM2_DURATION_LONG     15000 /* ms */
+
 #define TPM_REQ_HEADER_INITIALIZER(TAG, SIZE, ORD) \
     { \
         .tag = htobe16(TAG), \
@@ -452,7 +490,7 @@ static const char *get_name_for_bank(uint16_t hashAlg) {
     return NULL;
 }
 
-/* Give the name of a hash bank, return its algo identifer */
+/* Give the name of a hash bank, return its algo identifier */
 static uint16_t get_hashalg_by_bankname(const char *name) {
     size_t i;
 
@@ -474,7 +512,8 @@ static int swtpm_tpm2_shutdown(struct swtpm *self)
         .shutdownType = htobe16(TPM2_SU_CLEAR)
     };
 
-    return transfer(self, &req, sizeof(req), "TPM2_Shutdown", FALSE, NULL, NULL);
+    return transfer(self, &req, sizeof(req), "TPM2_Shutdown", FALSE,
+                    NULL, NULL, TPM2_DURATION_SHORT);
 }
 
 /* Get all available PCR banks */
@@ -502,15 +541,22 @@ static int swtpm_tpm2_get_all_pcr_banks(struct swtpm *self, gchar ***all_pcr_ban
     }
     ((struct tpm_req_header *)req)->size = htobe32(req_len);
 
-    ret = transfer(self, req, req_len, "TPM2_GetCapability", FALSE, tpmresp, &tpmresp_len);
+    ret = transfer(self, req, req_len, "TPM2_GetCapability", FALSE,
+                   tpmresp, &tpmresp_len, TPM2_DURATION_MEDIUM);
     if (ret != 0)
         return 1;
 
+    *all_pcr_banks = NULL;
+
     if (tpmresp_len < 17 + sizeof(count))
         goto err_too_short;
     memcpy(&count, &tpmresp[17], sizeof(count));
     count = be16toh(count);
 
+    /* unreasonable number of PCR banks ? */
+    if (count > 20)
+        goto err_num_pcrbanks;
+
     *all_pcr_banks = g_malloc0(sizeof(char *) * (count + 1));
 
     offset = 19;
@@ -539,8 +585,17 @@ static int swtpm_tpm2_get_all_pcr_banks(struct swtpm *self, gchar ***all_pcr_ban
     }
     return 0;
 
+err_num_pcrbanks:
+    logerr(self->logfile, "Unreasonable number of PCR banks (%u) returned.\n", count);
+    goto err_exit;
+
 err_too_short:
     logerr(self->logfile, "Response from TPM2_GetCapability is too short!\n");
+
+err_exit:
+    g_strfreev(*all_pcr_banks);
+    *all_pcr_banks = NULL;
+
     return 1;
 }
 
@@ -573,7 +628,7 @@ static int swtpm_tpm2_set_active_pcr_banks(struct swtpm *self, gchar **pcr_banks
                 break;
             }
         }
-        if (hashAlg != 0 && (activated_mask & (1 << j)) == 0) {
+        if (hashAlg != 0 && (activated_mask & ((uint64_t)1 << j)) == 0) {
             (*active)[count] = g_strdup(pcr_banks[idx]);
             len = concat(&pcrselects[pcrselects_len], sizeof(pcrselects) - pcrselects_len,
                          (unsigned char[]){AS2BE(hashAlg), 3, 0xff, 0xff, 0xff} , (size_t)6,
@@ -584,7 +639,7 @@ static int swtpm_tpm2_set_active_pcr_banks(struct swtpm *self, gchar **pcr_banks
             }
             pcrselects_len += len;
             count++;
-            activated_mask |= (1 << j);
+            activated_mask |= ((uint64_t)1 << j);
         }
     }
 
@@ -636,7 +691,8 @@ static int swtpm_tpm2_set_active_pcr_banks(struct swtpm *self, gchar **pcr_banks
     }
     ((struct tpm_req_header *)req)->size = htobe32(req_len);
 
-    ret = transfer(self, req, req_len, "TPM2_PCR_Allocate", FALSE, NULL, 0);
+    ret = transfer(self, req, req_len, "TPM2_PCR_Allocate", FALSE,
+                   NULL, NULL, TPM2_DURATION_SHORT);
     if (ret != 0)
         goto error;
 
@@ -649,6 +705,20 @@ error:
     return 1;
 }
 
+static int swtpm_tpm2_flushcontext(struct swtpm *self, uint32_t handle)
+{
+    struct tpm2_flushcontext_req {
+        struct tpm_req_header hdr;
+        uint32_t flushHandle;
+    } __attribute__((packed)) req = {
+        .hdr = TPM_REQ_HEADER_INITIALIZER(TPM2_ST_NO_SESSIONS, sizeof(req), TPM2_CC_FLUSHCONTEXT),
+        .flushHandle = htobe32(handle),
+    };
+
+    return transfer(self, &req, sizeof(req), "TPM2_FlushContext", FALSE,
+                    NULL, NULL, TPM2_DURATION_SHORT);
+}
+
 /* Make object at the curr_handler permanent with the perm_handle */
 static int swtpm_tpm2_evictcontrol(struct swtpm *self, uint32_t curr_handle, uint32_t perm_handle)
 {
@@ -668,7 +738,8 @@ static int swtpm_tpm2_evictcontrol(struct swtpm *self, uint32_t curr_handle, uin
         .persistentHandle = htobe32(perm_handle),
     };
 
-    return transfer(self, &req, sizeof(req), "TPM2_EvictControl", FALSE, NULL, 0);
+    return transfer(self, &req, sizeof(req), "TPM2_EvictControl", FALSE,
+                    NULL, NULL, TPM2_DURATION_SHORT);
 }
 
 /* Create an RSA EK */
@@ -840,7 +911,7 @@ static int swtpm_tpm2_createprimary_rsa(struct swtpm *self, uint32_t primaryhand
     ((struct tpm_req_header *)createprimary)->size = htobe32(createprimary_len);
 
     ret = transfer(self, createprimary, createprimary_len, "TPM2_CreatePrimary(RSA)", FALSE,
-                   tpmresp, &tpmresp_len);
+                   tpmresp, &tpmresp_len, TPM2_DURATION_LONG);
     if (ret != 0)
         return 1;
 
@@ -877,6 +948,7 @@ err_too_short:
 static int swtpm_tpm2_createprimary_ecc(struct swtpm *self, uint32_t primaryhandle, unsigned int keyflags,
                                         const unsigned char *symkeydata, size_t symkeydata_len,
                                         const unsigned char *authpolicy, size_t authpolicy_len,
+                                        const unsigned char *schemedata, size_t schemedata_len,
                                         unsigned short curveid, unsigned short hashalg,
                                         const unsigned char *nonce, size_t nonce_len,
                                         size_t off, uint32_t *curr_handle,
@@ -903,7 +975,7 @@ static int swtpm_tpm2_createprimary_ecc(struct swtpm *self, uint32_t primaryhand
                   }, (size_t)10,
                   authpolicy, authpolicy_len,
                   symkeydata, symkeydata_len,
-                  (unsigned char[]) {AS2BE(TPM2_ALG_NULL), AS2BE(curveid), AS2BE(TPM2_ALG_NULL)}, (size_t)6,
+                  schemedata, schemedata_len,
                   nonce, nonce_len,
                   nonce, nonce_len,
                   NULL);
@@ -937,7 +1009,7 @@ static int swtpm_tpm2_createprimary_ecc(struct swtpm *self, uint32_t primaryhand
     ((struct tpm_req_header *)createprimary)->size = htobe32(createprimary_len);
 
     ret = transfer(self, createprimary, createprimary_len, "TPM2_CreatePrimary(ECC)", FALSE,
-                   tpmresp, &tpmresp_len);
+                   tpmresp, &tpmresp_len, TPM2_DURATION_LONG);
     if (ret != 0)
         return 1;
     if (curr_handle) {
@@ -999,10 +1071,15 @@ static int swtpm_tpm2_createprimary_spk_ecc_nist_p384(struct swtpm *self,
     size_t authpolicy_len = sizeof(authpolicy);
     const unsigned char symkeydata[] = {AS2BE(TPM2_ALG_AES), AS2BE(256), AS2BE(TPM2_ALG_CFB)};
     size_t symkeydata_len = sizeof(symkeydata);
+    const unsigned char schemedata[] = {
+        AS2BE(TPM2_ALG_NULL), AS2BE(TPM2_ECC_NIST_P384), AS2BE(TPM2_ALG_NULL)
+    };
+    size_t schemedata_len = sizeof(schemedata);
     size_t off = 42;
 
     return swtpm_tpm2_createprimary_ecc(self, TPM2_RH_OWNER, keyflags, symkeydata, symkeydata_len,
-                                        authpolicy, authpolicy_len, TPM2_ECC_NIST_P384, TPM2_ALG_SHA384,
+                                        authpolicy, authpolicy_len, schemedata, schemedata_len,
+                                        TPM2_ECC_NIST_P384, TPM2_ALG_SHA384,
                                         NONCE_ECC_384, sizeof(NONCE_ECC_384), off, curr_handle,
                                         NULL, 0, NULL, NULL);
 }
@@ -1053,6 +1130,12 @@ static int swtpm_tpm2_create_spk(struct swtpm *self, gboolean isecc, unsigned in
         logit(self->logfile,
               "Successfully created storage primary key with handle 0x%x.\n", TPM2_SPK_HANDLE);
 
+    ret = swtpm_tpm2_flushcontext(self, curr_handle);
+    if (ret != 0) {
+        logerr(self->logfile, "Could not flush storage primary key.\n");
+        ret = 1;
+    }
+
     return ret;
 }
 
@@ -1068,6 +1151,10 @@ static int swtpm_tpm2_createprimary_ek_ecc_nist_p384(struct swtpm *self, gboolea
         0xCB, 0x1C, 0x0A, 0xD9, 0xBD, 0xE4, 0x19, 0xCA, 0xCB, 0x47, 0xBA, 0x09,
         0x69, 0x96, 0x46, 0x15, 0x0F, 0x9F, 0xC0, 0x00, 0xF3, 0xF8, 0x0E, 0x12
     };
+    const unsigned char schemedata[] = {
+        AS2BE(TPM2_ALG_NULL), AS2BE(TPM2_ECC_NIST_P384), AS2BE(TPM2_ALG_NULL)
+    };
+    size_t schemedata_len = sizeof(schemedata);
     size_t authpolicy_len = 48;
     unsigned char symkeydata[6];
     size_t symkeydata_len;
@@ -1103,8 +1190,11 @@ static int swtpm_tpm2_createprimary_ek_ecc_nist_p384(struct swtpm *self, gboolea
         off = 90;
     }
 
-    ret = swtpm_tpm2_createprimary_ecc(self, TPM2_RH_ENDORSEMENT, keyflags, symkeydata, symkeydata_len,
-                                       authpolicy, authpolicy_len, TPM2_ECC_NIST_P384, TPM2_ALG_SHA384,
+    ret = swtpm_tpm2_createprimary_ecc(self, TPM2_RH_ENDORSEMENT, keyflags,
+                                       symkeydata, symkeydata_len,
+                                       authpolicy, authpolicy_len,
+                                       schemedata, schemedata_len,
+                                       TPM2_ECC_NIST_P384, TPM2_ALG_SHA384,
                                        NONCE_EMPTY, sizeof(NONCE_EMPTY), off, curr_handle,
                                        ektemplate, ektemplate_len, ekparam, key_description);
     if (ret != 0)
@@ -1213,7 +1303,8 @@ static int swtpm_tpm2_nvdefinespace(struct swtpm *self, uint32_t nvindex, uint32
 
     ((struct tpm_req_header *)req)->size = htobe32(req_len);
 
-    return transfer(self, req, req_len, "TPM2_NV_DefineSpace", FALSE, NULL, 0);
+    return transfer(self, req, req_len, "TPM2_NV_DefineSpace", FALSE,
+                    NULL, NULL, TPM2_DURATION_SHORT);
 }
 
 /* Write the data into the given NVIndex */
@@ -1247,7 +1338,8 @@ static int swtpm_tpm2_nv_write(struct swtpm *self, uint32_t nvindex,
         }
         ((struct tpm_req_header *)req)->size = htobe32(req_len);
 
-        ret = transfer(self, req, req_len, "TPM2_NV_Write", FALSE, NULL, 0);
+        ret = transfer(self, req, req_len, "TPM2_NV_Write", FALSE,
+                       NULL, NULL, TPM2_DURATION_SHORT);
         if (ret != 0)
             return 1;
 
@@ -1277,23 +1369,24 @@ static int swtpm_tpm2_nv_writelock(struct swtpm *self, uint32_t nvindex)
 
     ((struct tpm_req_header *)req)->size = htobe32(req_len);
 
-    return transfer(self, req, req_len, "TPM2_NV_WriteLock", FALSE, NULL, 0);
+    return transfer(self, req, req_len, "TPM2_NV_WriteLock", FALSE,
+                    NULL, NULL, TPM2_DURATION_SHORT);
 }
 
 static int swtpm_tpm2_write_nvram(struct swtpm *self, uint32_t nvindex, uint32_t nvindexattrs,
                                   const unsigned char *data, size_t data_len, gboolean lock_nvram,
-                                  const char *purpose)
+                                  const char *certtype)
 {
     int ret = swtpm_tpm2_nvdefinespace(self, nvindex, nvindexattrs, data_len);
     if (ret != 0) {
-        logerr(self->logfile, "Could not create NVRAM area 0x%x for %s.\n", nvindex, purpose);
+        logerr(self->logfile, "Could not create NVRAM area 0x%x for %s.\n", nvindex, certtype);
         return 1;
     }
 
     ret = swtpm_tpm2_nv_write(self, nvindex, data, data_len);
     if (ret != 0) {
         logerr(self->logfile,
-               "Could not write %s into NVRAM area 0x%x.\n", purpose, nvindex);
+               "Could not write %s into NVRAM area 0x%x.\n", certtype, nvindex);
         return 1;
     }
 
@@ -1308,6 +1401,24 @@ static int swtpm_tpm2_write_nvram(struct swtpm *self, uint32_t nvindex, uint32_t
     return 0;
 }
 
+static int swtpm_tpm2_write_cert_nvram(struct swtpm *self, uint32_t nvindex,
+                                       uint32_t nvindexattrs,
+                                       const unsigned char *data, size_t data_len,
+                                       gboolean lock_nvram, const char *keytype,
+                                       const char *certtype)
+{
+    int ret;
+
+    ret = swtpm_tpm2_write_nvram(self, nvindex, nvindexattrs, data, data_len, lock_nvram,
+                                 certtype);
+    if (ret == 0)
+        logit(self->logfile,
+              "Successfully created NVRAM area 0x%x for %s%s.\n",
+              nvindex, keytype, certtype);
+
+    return ret;
+}
+
 /* Write the platform certificate into an NVRAM area */
 static int swtpm_tpm2_write_ek_cert_nvram(struct swtpm *self, gboolean isecc,
                                            unsigned int rsa_keysize, gboolean lock_nvram,
@@ -1322,30 +1433,20 @@ static int swtpm_tpm2_write_ek_cert_nvram(struct swtpm *self, gboolean isecc,
             TPMA_NV_PPWRITE |
             TPMA_NV_NO_DA |
             TPMA_NV_WRITEDEFINE;
-    int ret;
 
     if (!isecc) {
         if (rsa_keysize == 2048)
             nvindex = TPM2_NV_INDEX_RSA2048_EKCERT;
         else if (rsa_keysize == 3072)
             nvindex = TPM2_NV_INDEX_RSA3072_HI_EKCERT;
-        keytype = g_strdup_printf("RSA %d", rsa_keysize);
+        keytype = g_strdup_printf("RSA %d ", rsa_keysize);
     } else {
         nvindex = TPM2_NV_INDEX_ECC_SECP384R1_HI_EKCERT;
-        keytype = g_strdup("ECC");
+        keytype = g_strdup("ECC ");
     }
 
-    ret = swtpm_tpm2_write_nvram(self, nvindex, nvindexattrs, data, data_len, lock_nvram,
-                                 "EK Certificate");
-    if (ret == 0)
-        logit(self->logfile,
-              "Successfully created NVRAM area 0x%x for %s EK certificate.\n",
-              nvindex, keytype);
-    else
-        logerr(self->logfile,
-               "Could not create NVRAM area 0x%x for %s EK certificate.\n",
-               nvindex, keytype);
-    return ret;
+    return swtpm_tpm2_write_cert_nvram(self, nvindex, nvindexattrs, data, data_len,
+                                       lock_nvram, keytype, "EK certificate");
 }
 
 static int swtpm_tpm2_write_platform_cert_nvram(struct swtpm *self, gboolean lock_nvram,
@@ -1359,18 +1460,9 @@ static int swtpm_tpm2_write_platform_cert_nvram(struct swtpm *self, gboolean loc
             TPMA_NV_PPWRITE |
             TPMA_NV_NO_DA |
             TPMA_NV_WRITEDEFINE;
-    int ret;
 
-    ret = swtpm_tpm2_write_nvram(self, nvindex, nvindexattrs, data, data_len, lock_nvram,
-                                 "Platform Certificate");
-    if (ret == 0)
-        logit(self->logfile,
-              "Successfully created NVRAM area 0x%x for platform certificate.\n", nvindex);
-    else
-        logerr(self->logfile,
-               "Could not create NVRAM area 0x%x for platform certificate.\n", nvindex);
-
-    return ret;
+    return swtpm_tpm2_write_cert_nvram(self, nvindex, nvindexattrs, data, data_len,
+                                       lock_nvram, "", "platform certificate");
 }
 
 static const struct swtpm2_ops swtpm_tpm2_ops = {
@@ -1435,6 +1527,9 @@ static const struct swtpm2_ops swtpm_tpm2_ops = {
 
 #define TPM_KH_EK    0x40000006
 
+#define TPM_DURATION_SHORT     2000  /* ms */
+#define TPM_DURATION_MEDIUM    7500  /* ms */
+#define TPM_DURATION_LONG     15000  /* ms */
 
 static int swtpm_tpm12_tsc_physicalpresence(struct swtpm *self, uint16_t physicalpresence)
 {
@@ -1446,14 +1541,17 @@ static int swtpm_tpm12_tsc_physicalpresence(struct swtpm *self, uint16_t physica
         .pp = htobe16(physicalpresence),
     };
 
-    return transfer(self, &req, sizeof(req), "TSC_PhysicalPresence", FALSE, NULL, NULL);
+    /* use medium duration to avoid t/o on busy system */
+    return transfer(self, &req, sizeof(req), "TSC_PhysicalPresence", FALSE,
+                    NULL, NULL, TPM_DURATION_MEDIUM);
 }
 
 static int swtpm_tpm12_physical_enable(struct swtpm *self)
 {
     struct tpm_req_header req = TPM_REQ_HEADER_INITIALIZER(TPM_TAG_RQU_COMMAND, sizeof(req), TPM_ORD_PHYSICAL_ENABLE);
 
-    return transfer(self, &req, sizeof(req), "TPM_PhysicalEnable", FALSE, NULL, NULL);
+    return transfer(self, &req, sizeof(req), "TPM_PhysicalEnable", FALSE,
+                    NULL, NULL, TPM_DURATION_SHORT);
 }
 
 static int swtpm_tpm12_physical_set_deactivated(struct swtpm *self, uint8_t state)
@@ -1466,7 +1564,8 @@ static int swtpm_tpm12_physical_set_deactivated(struct swtpm *self, uint8_t stat
         .state = state,
     };
 
-    return transfer(self, &req, sizeof(req), "TSC_PhysicalSetDeactivated", FALSE, NULL, NULL);
+    return transfer(self, &req, sizeof(req), "TSC_PhysicalSetDeactivated", FALSE,
+                    NULL, NULL, TPM_DURATION_SHORT);
 }
 
 /* Initialize the TPM1.2 */
@@ -1495,7 +1594,8 @@ static int swptm_tpm12_create_endorsement_keypair(struct swtpm *self,
     uint32_t length;
     int ret;
 
-    ret = transfer(self, &req, sizeof(req), "TPM_CreateEndorsementKeyPair", FALSE, &tpmresp, &tpmresp_len);
+    ret = transfer(self, &req, sizeof(req), "TPM_CreateEndorsementKeyPair", FALSE,
+                   &tpmresp, &tpmresp_len, TPM_DURATION_LONG);
     if (ret != 0)
         return 1;
 
@@ -1529,7 +1629,8 @@ static int swtpm_tpm12_oiap(struct swtpm *self, uint32_t *authhandle, unsigned c
     size_t tpmresp_len = sizeof(tpmresp);
     int ret;
 
-    ret = transfer(self, &req, sizeof(req), "TPM_OIAP", FALSE, &tpmresp, &tpmresp_len);
+    ret = transfer(self, &req, sizeof(req), "TPM_OIAP", FALSE,
+                   &tpmresp, &tpmresp_len, TPM_DURATION_SHORT);
     if (ret != 0)
         return ret;
 
@@ -1653,7 +1754,7 @@ static int swtpm_tpm12_take_ownership(struct swtpm *self, const unsigned char ow
         EVP_PKEY_CTX_set_rsa_padding(ctx, RSA_PKCS1_OAEP_PADDING) < 1 ||
         EVP_PKEY_CTX_set_rsa_mgf1_md(ctx, sha1) < 1 ||
         EVP_PKEY_CTX_set_rsa_oaep_md(ctx, sha1) < 1 ||
-        EVP_PKEY_CTX_set0_rsa_oaep_label(ctx, strdup("TCPA"), 4) < 1 ||
+        EVP_PKEY_CTX_set0_rsa_oaep_label(ctx, g_strdup("TCPA"), 4) < 1 ||
         EVP_PKEY_encrypt(ctx, enc_owner_auth, &enc_owner_auth_len,
                          ownerpass_digest, SHA_DIGEST_LENGTH) < 1||
         EVP_PKEY_encrypt(ctx, enc_srk_auth, &enc_srk_auth_len,
@@ -1752,7 +1853,8 @@ static int swtpm_tpm12_take_ownership(struct swtpm *self, const unsigned char ow
     trh = (struct tpm_req_header *)req; /* old gcc type-punned pointer */
     trh->size = htobe32(req_len);
 
-    ret = transfer(self, req, req_len, "TPM_TakeOwnership", FALSE, NULL, 0);
+    ret = transfer(self, req, req_len, "TPM_TakeOwnership", FALSE,
+                   NULL, NULL, TPM_DURATION_LONG);
 
 error:
     EVP_PKEY_free(pkey);
@@ -1828,7 +1930,8 @@ static int swtpm_tpm12_nv_define_space(struct swtpm *self, uint32_t nvindex,
 
     ((struct tpm_req_header *)req)->size = htobe32(req_len);
 
-    return transfer(self, req, req_len, "TPM_NV_DefineSpace", FALSE, NULL, 0);
+    return transfer(self, req, req_len, "TPM_NV_DefineSpace", FALSE,
+                    NULL, NULL, TPM_DURATION_SHORT);
 }
 
 static int swtpm_tpm12_nv_write_value(struct swtpm *self, uint32_t nvindex,
@@ -1850,7 +1953,8 @@ static int swtpm_tpm12_nv_write_value(struct swtpm *self, uint32_t nvindex,
 
     ((struct tpm_req_header *)req)->size = htobe32(req_len);
 
-    return transfer(self, req, req_len, "TPM_NV_DefineSpace", FALSE, NULL, 0);
+    return transfer(self, req, req_len, "TPM_NV_DefineSpace", FALSE,
+                    NULL, NULL, TPM_DURATION_SHORT);
 }
 
 /* Write the EK Certificate into NVRAM */
