@@ -188,7 +188,7 @@ ssize_t memconcat(unsigned char **buffer, ...)
 }
 
 /* Print an input buffer as hex number */
-gchar *print_as_hex(unsigned char *input, size_t input_len)
+gchar *print_as_hex(const unsigned char *input, size_t input_len)
 {
     gchar *out = g_malloc(input_len * 2 + 1);
     size_t i;
@@ -377,7 +377,7 @@ error:
     return -1;
 }
 
-/* replace occurences of 'torep' with 'rep' in a string 'in' */
+/* replace occurrences of 'torep' with 'rep' in a string 'in' */
 gchar *str_replace(const char *in, const char *torep, const char *rep)
 {
     char *res;
