@@ -10,9 +10,24 @@
 #ifndef SWTPM_UTILS_H
 #define SWTPM_UTILS_H
 
+#include <pwd.h>
+
 #include <glib.h>
 
-#define min(X,Y) (X) < (Y) ? (X) : (Y)
+#define min(X,Y) ((X) < (Y) ? (X) : (Y))
+#define ARRAY_LEN(a) (sizeof(a) / sizeof((a)[0]))
+
+#define SWTPM_CLOSE(FD)	\
+    if ((FD) >= 0) {	\
+        close((FD));	\
+        (FD) = -1;	\
+    }
+
+#define SWTPM_G_FREE(var)	\
+    do {		\
+        g_free(var);	\
+        var = NULL;	\
+    } while(0)
 
 extern gchar *gl_LOGFILE;
 
@@ -27,7 +42,7 @@ ssize_t concat(unsigned char *buf, size_t buflen, ...);
 ssize_t memconcat(unsigned char **buffer, ...);
 
 gchar *resolve_string(const gchar *inp);
-gchar *print_as_hex(unsigned char *input, size_t input_len);
+gchar *print_as_hex(const unsigned char *input, size_t input_len);
 gchar **split_cmdline(const gchar *cmdline);
 
 int read_file(const gchar *filename, gchar **buffer, gsize *buffer_len);
@@ -38,4 +53,6 @@ int write_to_tempfile(gchar **filename, const unsigned char *data, size_t data_l
 
 gchar *str_replace(const char *in, const char *torep, const char *rep);
 
+int check_directory_access(const gchar *directory, int mode, const struct passwd *curr_user);
+
 #endif /* SWTPM_UTILS_H */
